import os
import warnings
from typing import Any, Dict, Optional

import hydra
from lightning import Fabric
from lightning.fabric.loggers.logger import Logger
from lightning.fabric.plugins.collectives import TorchCollective
from lightning.fabric.utilities.cloud_io import _is_dir, get_filesystem


def get_logger(fabric: Fabric, cfg: Dict[str, Any]) -> Optional[Logger]:
    # Set logger only on rank-0 but share the logger directory: since we don't know
    # what is happening during the `fabric.save()` method, at least we assure that all
    # ranks save under the same named folder.
    # As a plus, rank-0 sets the time uniquely for everyone
    logger = None
    if fabric.is_global_zero and cfg.metric.log_level > 0:
        if "tensorboard" in cfg.metric.logger._target_.lower():
            root_dir = os.path.join("logs", "runs", cfg.root_dir)
            if root_dir != cfg.metric.logger.root_dir:
                warnings.warn(
                    "The specified root directory for the TensorBoardLogger is different from the experiment one, "
                    "so the logger one will be ignored and replaced with the experiment root directory",
                    UserWarning,
                )
            if cfg.run_name != cfg.metric.logger.name:
                warnings.warn(
                    "The specified name for the TensorBoardLogger is different from the `run_name` of the experiment, "
                    "so the logger one will be ignored and replaced with the experiment `run_name`",
                    UserWarning,
                )
            cfg.metric.logger.root_dir = root_dir
            cfg.metric.logger.name = cfg.run_name
        logger = hydra.utils.instantiate(cfg.metric.logger, _convert_="all")
    return logger


def get_log_dir(fabric: Fabric, root_dir: str, run_name: str, share: bool = True) -> str:
    """Return and, if necessary, create the log directory. If there are more than one processes,
    the rank-0 process shares the directory to the others (if the `share` parameter is set to `True`).

    Args:
        fabric (Fabric): the fabric instance.
        root_dir (str): the root directory of the experiment.
        run_name (str): the name of the experiment.
        share (bool): whether or not to share the `log_dir` among processes.

    Returns:
        The log directory of the experiment.
    """
    world_collective = TorchCollective()
    if fabric.world_size > 1 and share:
        world_collective.setup()
        world_collective.create_group()
    if fabric.is_global_zero:
        # If the logger was instantiated, then take the log_dir from it
        if len(fabric.loggers) > 0 and fabric.logger.log_dir is not None:
            log_dir = fabric.logger.log_dir
        else:
            # Otherwise the rank-zero process creates the log_dir
            save_dir = os.path.join("logs", "runs", root_dir, run_name)
            fs = get_filesystem(root_dir)
            try:
                listdir_info = fs.listdir(save_dir)
                existing_versions = []
                for listing in listdir_info:
                    d = listing["name"]
                    bn = os.path.basename(d)
                    if _is_dir(fs, d) and bn.startswith("version_"):
                        dir_ver = bn.split("_")[1].replace("/", "")
                        existing_versions.append(int(dir_ver))
                if len(existing_versions) == 0:
                    version = 0
                else:
                    version = max(existing_versions) + 1
                log_dir = os.path.join(save_dir, f"version_{version}")
            except OSError:
                warnings.warn("Missing logger folder: %s" % save_dir, UserWarning)
                log_dir = os.path.join(save_dir, f"version_{0}")

            os.makedirs(log_dir, exist_ok=True)
        if fabric.world_size > 1 and share:
            world_collective.broadcast_object_list([log_dir], src=0)
    else:
        data = [None]
        world_collective.broadcast_object_list(data, src=0)
        log_dir = data[0]
    return log_dir



import copy
import os
import warnings
from datetime import timedelta
from math import prod
from typing import Any, Dict

import gymnasium as gym
import hydra
import numpy as np
import torch
from lightning.fabric import Fabric
from lightning.fabric.plugins.collectives import TorchCollective
from lightning.fabric.plugins.collectives.collective import CollectibleGroup
from lightning.fabric.strategies import DDPStrategy
from torch.utils.data.sampler import BatchSampler
from torchmetrics import SumMetric

from sheeprl.algos.sac.agent import SACActor, SACAgent, SACCritic, build_agent
from sheeprl.algos.sac.sac import train
from sheeprl.algos.sac.utils import test
from sheeprl.data.buffers import ReplayBuffer
from sheeprl.utils.env import make_env
from sheeprl.utils.logger import get_log_dir
from sheeprl.utils.metric import MetricAggregator
from sheeprl.utils.registry import register_algorithm
from sheeprl.utils.timer import timer
from sheeprl.utils.utils import save_configs


@torch.no_grad()
def player(
    fabric: Fabric, cfg: Dict[str, Any], world_collective: TorchCollective, player_trainer_collective: TorchCollective
):
    log_dir = get_log_dir(fabric, cfg.root_dir, cfg.run_name, False)
    rank = fabric.global_rank
    device = fabric.device
    fabric.seed_everything(cfg.seed)
    torch.backends.cudnn.deterministic = cfg.torch_deterministic

    # Resume from checkpoint
    if cfg.checkpoint.resume_from:
        state = fabric.load(cfg.checkpoint.resume_from)

    if len(cfg.algo.cnn_keys.encoder) > 0:
        warnings.warn("SAC algorithm cannot allow to use images as observations, the CNN keys will be ignored")
        cfg.algo.cnn_keys.encoder = []

    # Environment setup
    vectorized_env = gym.vector.SyncVectorEnv if cfg.env.sync_env else gym.vector.AsyncVectorEnv
    envs = vectorized_env(
        [
            make_env(
                cfg,
                cfg.seed + rank * cfg.env.num_envs + i,
                rank * cfg.env.num_envs,
                log_dir if rank == 0 else None,
                "train",
                vector_env_idx=i,
            )
            for i in range(cfg.env.num_envs)
        ]
    )
    action_space = envs.single_action_space
    observation_space = envs.single_observation_space
    if not isinstance(action_space, gym.spaces.Box):
        raise ValueError("Only continuous action space is supported for the SAC agent")
    if not isinstance(observation_space, gym.spaces.Dict):
        raise RuntimeError(f"Unexpected observation type, should be of type Dict, got: {observation_space}")
    if len(cfg.algo.mlp_keys.encoder) == 0:
        raise RuntimeError("You should specify at least one MLP key for the encoder: `mlp_keys.encoder=[state]`")
    for k in cfg.algo.mlp_keys.encoder:
        if len(observation_space[k].shape) > 1:
            raise ValueError(
                "Only environments with vector-only observations are supported by the SAC agent. "
                f"Provided environment: {cfg.env.id}"
            )
    if cfg.metric.log_level > 0:
        fabric.print("Encoder MLP keys:", cfg.algo.mlp_keys.encoder)

    if fabric.is_global_zero:
        save_configs(cfg, log_dir)

    # Send (possibly updated, by the make_env method for example) cfg to the trainers
    if cfg.checkpoint.resume_from:
        cfg.algo.per_rank_batch_size = state["batch_size"] // fabric.world_size
    cfg.checkpoint.log_dir = log_dir
    world_collective.broadcast_object_list([cfg], src=0)

    # Define the agent and the optimizer and setup them with Fabric
    act_dim = prod(action_space.shape)
    obs_dim = sum([prod(observation_space[k].shape) for k in cfg.algo.mlp_keys.encoder])
    actor = SACActor(
        observation_dim=obs_dim,
        action_dim=act_dim,
        distribution_cfg=cfg.distribution,
        hidden_size=cfg.algo.actor.hidden_size,
        action_low=action_space.low,
        action_high=action_space.high,
    ).to(device)
    flattened_parameters = torch.empty_like(
        torch.nn.utils.convert_parameters.parameters_to_vector(actor.parameters()), device=device
    )

    # Receive the first weights from the rank-1, a.k.a. the first of the trainers
    # In this way we are sure that before the first iteration everyone starts with the same parameters
    player_trainer_collective.broadcast(flattened_parameters, src=1)
    torch.nn.utils.convert_parameters.vector_to_parameters(flattened_parameters, actor.parameters())

    # Metrics
    aggregator = None
    if not MetricAggregator.disabled:
        aggregator: MetricAggregator = hydra.utils.instantiate(cfg.metric.aggregator, _convert_="all").to(device)

    # Local data
    buffer_size = cfg.buffer.size // cfg.env.num_envs if not cfg.dry_run else 1
    rb = ReplayBuffer(
        buffer_size,
        cfg.env.num_envs,
        memmap=cfg.buffer.memmap,
        memmap_dir=os.path.join(log_dir, "memmap_buffer", f"rank_{fabric.global_rank}"),
    )
    if cfg.checkpoint.resume_from and cfg.buffer.checkpoint:
        if isinstance(state["rb"], ReplayBuffer):
            rb = state["rb"]
        else:
            raise RuntimeError(
                "The replay buffer in the configs must be of type "
                f"`sheeprl.data.buffers.ReplayBuffer`, got {type(state['rb'])}."
            )

    # Global variables
    first_info_sent = False
    start_step = (
        # + 1 because the checkpoint is at the end of the update step
        # (when resuming from a checkpoint, the update at the checkpoint
        # is ended and you have to start with the next one)
        state["update"] + 1
        if cfg.checkpoint.resume_from
        else 1
    )
    policy_step = state["update"] * cfg.env.num_envs if cfg.checkpoint.resume_from else 0
    last_log = state["last_log"] if cfg.checkpoint.resume_from else 0
    last_checkpoint = state["last_checkpoint"] if cfg.checkpoint.resume_from else 0
    policy_steps_per_update = int(cfg.env.num_envs)
    num_updates = int(cfg.algo.total_steps // policy_steps_per_update) if not cfg.dry_run else 1
    learning_starts = cfg.algo.learning_starts // policy_steps_per_update if not cfg.dry_run else 0
    if cfg.checkpoint.resume_from and not cfg.buffer.checkpoint:
        learning_starts += start_step

    # Warning for log and checkpoint every
    if cfg.metric.log_level > 0 and cfg.metric.log_every % policy_steps_per_update != 0:
        warnings.warn(
            f"The metric.log_every parameter ({cfg.metric.log_every}) is not a multiple of the "
            f"policy_steps_per_update value ({policy_steps_per_update}), so "
            "the metrics will be logged at the nearest greater multiple of the "
            "policy_steps_per_update value."
        )
    if cfg.checkpoint.every % policy_steps_per_update != 0:
        warnings.warn(
            f"The checkpoint.every parameter ({cfg.checkpoint.every}) is not a multiple of the "
            f"policy_steps_per_update value ({policy_steps_per_update}), so "
            "the checkpoint will be saved at the nearest greater multiple of the "
            "policy_steps_per_update value."
        )

    step_data = {}
    # Get the first environment observation and start the optimization
    obs = envs.reset(seed=cfg.seed)[0]
    obs = np.concatenate([obs[k] for k in cfg.algo.mlp_keys.encoder], axis=-1)

    for update in range(start_step, num_updates + 1):
        policy_step += cfg.env.num_envs

        # Measure environment interaction time: this considers both the model forward
        # to get the action given the observation and the time taken into the environment
        with timer("Time/env_interaction_time", SumMetric, sync_on_compute=False):
            if update <= learning_starts:
                actions = envs.action_space.sample()
            else:
                # Sample an action given the observation received by the environment
                with torch.no_grad():
                    torch_obs = torch.as_tensor(obs, dtype=torch.float32, device=device)
                    actions, _ = actor(torch_obs)
                    actions = actions.cpu().numpy()
            next_obs, rewards, dones, truncated, infos = envs.step(actions)
            next_obs = np.concatenate([next_obs[k] for k in cfg.algo.mlp_keys.encoder], axis=-1)
            dones = np.logical_or(dones, truncated).reshape(cfg.env.num_envs, -1).astype(np.uint8)
            rewards = rewards.reshape(cfg.env.num_envs, -1)

        if cfg.metric.log_level > 0 and "final_info" in infos:
            for i, agent_ep_info in enumerate(infos["final_info"]):
                if agent_ep_info is not None:
                    ep_rew = agent_ep_info["episode"]["r"]
                    ep_len = agent_ep_info["episode"]["l"]
                    if aggregator and not aggregator.disabled:
                        aggregator.update("Rewards/rew_avg", ep_rew)
                        aggregator.update("Game/ep_len_avg", ep_len)
                    fabric.print(f"Rank-0: policy_step={policy_step}, reward_env_{i}={ep_rew[-1]}")

        # Save the real next observation
        real_next_obs = copy.deepcopy(next_obs)
        if "final_observation" in infos:
            for idx, final_obs in enumerate(infos["final_observation"]):
                if final_obs is not None:
                    real_next_obs[idx] = np.concatenate(
                        [v for k, v in final_obs.items() if k in cfg.algo.mlp_keys.encoder], axis=-1
                    )

        step_data["dones"] = dones[np.newaxis]
        step_data["actions"] = actions[np.newaxis]
        step_data["observations"] = obs[np.newaxis]
        if not cfg.buffer.sample_next_obs:
            step_data["next_observations"] = real_next_obs[np.newaxis]
        step_data["rewards"] = rewards[np.newaxis]
        rb.add(step_data, validate_args=cfg.buffer.validate_args)

        # next_obs becomes the new obs
        obs = next_obs

        # Send data to the training agents
        if update >= learning_starts:
            # Send local info to the trainers
            if not first_info_sent:
                world_collective.broadcast_object_list(
                    [{"update": update, "last_log": last_log, "last_checkpoint": last_checkpoint}], src=0
                )
                first_info_sent = True

            # Sample data to be sent to the trainers
            training_steps = learning_starts if update == learning_starts else 1
            sample = rb.sample_tensors(
                batch_size=training_steps
                * cfg.algo.per_rank_gradient_steps
                * cfg.algo.per_rank_batch_size
                * (fabric.world_size - 1),
                sample_next_obs=cfg.buffer.sample_next_obs,
                dtype=None,
                device=device,
                from_numpy=cfg.buffer.from_numpy,
            )
            # chunks = {k1: [k1_chunk_1, k1_chunk_2, ...], k2: [k2_chunk_1, k2_chunk_2, ...]}
            chunks = {
                k: v.float().split(training_steps * cfg.algo.per_rank_gradient_steps * cfg.algo.per_rank_batch_size)
                for k, v in sample.items()
            }
            # chunks = [{k1: k1_chunk_1, k2: k2_chunk_1}, {k1: k1_chunk_2, k2: k2_chunk_2}, ...]
            chunks = [{k: v[i] for k, v in chunks.items()} for i in range(len(chunks[next(iter(chunks.keys()))]))]
            world_collective.scatter_object_list([None], [None] + chunks, src=0)

            # Wait the trainers to finish
            player_trainer_collective.broadcast(flattened_parameters, src=1)

            # Convert back the parameters
            torch.nn.utils.convert_parameters.vector_to_parameters(flattened_parameters, actor.parameters())

            # Logs trainers-only metrics
            if cfg.metric.log_level > 0 and policy_step - last_log >= cfg.metric.log_every:
                # Gather metrics from the trainers
                metrics = [None]
                player_trainer_collective.broadcast_object_list(metrics, src=1)

                # Log metrics
                fabric.log_dict(metrics[0], policy_step)

        # Logs player-only metrics
        if cfg.metric.log_level > 0 and policy_step - last_log >= cfg.metric.log_every:
            if aggregator and not aggregator.disabled:
                fabric.log_dict(aggregator.compute(), policy_step)
                aggregator.reset()

            # Sync timers
            if not timer.disabled:
                timer_metrics = timer.compute()
                fabric.log(
                    "Time/sps_env_interaction",
                    ((policy_step - last_log) * cfg.env.action_repeat) / timer_metrics["Time/env_interaction_time"],
                    policy_step,
                )
                timer.reset()

            # Reset counters
            last_log = policy_step

        # Checkpoint model
        if (
            update >= learning_starts  # otherwise the processes end up deadlocked
            and cfg.checkpoint.every > 0
            and policy_step - last_checkpoint >= cfg.checkpoint.every
        ):
            last_checkpoint = policy_step
            ckpt_path = log_dir + f"/checkpoint/ckpt_{policy_step}_{fabric.global_rank}.ckpt"
            fabric.call(
                "on_checkpoint_player",
                fabric=fabric,
                player_trainer_collective=player_trainer_collective,
                ckpt_path=ckpt_path,
                replay_buffer=rb if cfg.buffer.checkpoint else None,
            )

    world_collective.scatter_object_list([None], [None] + [-1] * (world_collective.world_size - 1), src=0)

    # Last Checkpoint
    if cfg.checkpoint.save_last:
        ckpt_path = log_dir + f"/checkpoint/ckpt_{policy_step}_{fabric.global_rank}.ckpt"
        fabric.call(
            "on_checkpoint_player",
            fabric=fabric,
            player_trainer_collective=player_trainer_collective,
            ckpt_path=ckpt_path,
            replay_buffer=rb if cfg.buffer.checkpoint else None,
        )

    envs.close()
    if fabric.is_global_zero and cfg.algo.run_test:
        test(actor, fabric, cfg, log_dir)

    if not cfg.model_manager.disabled and fabric.is_global_zero:
        from sheeprl.algos.sac.utils import log_models
        from sheeprl.utils.mlflow import register_model

        critics = [
            SACCritic(observation_dim=obs_dim + act_dim, hidden_size=cfg.algo.critic.hidden_size, num_critics=1)
            for _ in range(cfg.algo.critic.n)
        ]
        target_entropy = -act_dim
        agent = SACAgent(
            actor, critics, target_entropy, alpha=cfg.algo.alpha.alpha, tau=cfg.algo.tau, device=fabric.device
        )
        flattened_parameters = torch.empty_like(
            torch.nn.utils.convert_parameters.parameters_to_vector(agent.parameters()), device=device
        )
        player_trainer_collective.broadcast(flattened_parameters, src=1)
        torch.nn.utils.convert_parameters.vector_to_parameters(flattened_parameters, agent.parameters())
        register_model(fabric, log_models, cfg, {"agent": agent})


def trainer(
    world_collective: TorchCollective,
    player_trainer_collective: TorchCollective,
    optimization_pg: CollectibleGroup,
    cfg: Dict[str, Any],
):
    global_rank = world_collective.rank
    group_world_size = world_collective.world_size - 1

    # Initialize Fabric
    cfg.fabric.pop("loggers", None)
    cfg.fabric.pop("strategy", None)
    fabric: Fabric = hydra.utils.instantiate(
        cfg.fabric, strategy=DDPStrategy(process_group=optimization_pg), _convert_="all"
    )
    fabric.launch()
    device = fabric.device
    fabric.seed_everything(cfg.seed)
    torch.backends.cudnn.deterministic = cfg.torch_deterministic

    # Resume from checkpoint
    if cfg.checkpoint.resume_from:
        state = fabric.load(cfg.checkpoint.resume_from)

    # Receive (possibly updated, by the make_env method for example) cfg from the player
    data = [None]
    world_collective.broadcast_object_list(data, src=0)
    cfg: Dict[str, Any] = data[0]

    # Environment setup
    vectorized_env = gym.vector.SyncVectorEnv if cfg.env.sync_env else gym.vector.AsyncVectorEnv
    envs = vectorized_env([make_env(cfg, 0, 0, None)])
    assert isinstance(envs.single_action_space, gym.spaces.Box), "only continuous action space is supported"

    # Define the agent and the optimizer and setup them with Fabric
    agent = build_agent(
        fabric,
        cfg,
        envs.single_observation_space,
        envs.single_action_space,
        state["agent"] if cfg.checkpoint.resume_from else None,
    )

    # Optimizers
    qf_optimizer = hydra.utils.instantiate(cfg.algo.critic.optimizer, params=agent.qfs.parameters(), _convert_="all")
    actor_optimizer = hydra.utils.instantiate(
        cfg.algo.actor.optimizer, params=agent.actor.parameters(), _convert_="all"
    )
    alpha_optimizer = hydra.utils.instantiate(cfg.algo.alpha.optimizer, params=[agent.log_alpha], _convert_="all")
    if cfg.checkpoint.resume_from:
        qf_optimizer.load_state_dict(state["qf_optimizer"])
        actor_optimizer.load_state_dict(state["actor_optimizer"])
        alpha_optimizer.load_state_dict(state["alpha_optimizer"])
    qf_optimizer, actor_optimizer, alpha_optimizer = fabric.setup_optimizers(
        qf_optimizer, actor_optimizer, alpha_optimizer
    )

    # Send weights to rank-0, a.k.a. the player
    if global_rank == 1:
        player_trainer_collective.broadcast(
            torch.nn.utils.convert_parameters.parameters_to_vector(agent.actor.parameters()), src=1
        )

    # Metrics
    aggregator = None
    if not MetricAggregator.disabled:
        aggregator: MetricAggregator = hydra.utils.instantiate(cfg.metric.aggregator, _convert_="all").to(device)

    # Receive data from player reagrding the:
    # * update
    # * last_log
    # * last_checkpoint
    data = [None]
    world_collective.broadcast_object_list(data, src=0)
    update = data[0]["update"]
    last_log = data[0]["last_log"]
    last_checkpoint = data[0]["last_checkpoint"]

    # Start training
    train_step = 0
    last_train = 0
    policy_steps_per_update = cfg.env.num_envs
    policy_step = update * policy_steps_per_update
    while True:
        # Wait for data
        data = [None]
        world_collective.scatter_object_list(data, [None for _ in range(world_collective.world_size)], src=0)
        data = data[0]
        if not isinstance(data, dict) and data == -1:
            # Last Checkpoint
            if cfg.checkpoint.save_last:
                state = {
                    "agent": agent.state_dict(),
                    "qf_optimizer": qf_optimizer.state_dict(),
                    "actor_optimizer": actor_optimizer.state_dict(),
                    "alpha_optimizer": alpha_optimizer.state_dict(),
                    "update": update,
                    "batch_size": cfg.algo.per_rank_batch_size * (world_collective.world_size - 1),
                    "last_log": last_log,
                    "last_checkpoint": last_checkpoint,
                }
                ckpt_path = cfg.checkpoint.log_dir + f"/checkpoint/ckpt_{policy_step}_{fabric.global_rank}.ckpt"
                fabric.call(
                    "on_checkpoint_trainer",
                    fabric=fabric,
                    player_trainer_collective=player_trainer_collective,
                    ckpt_path=ckpt_path,
                    state=state,
                )
            if not cfg.model_manager.disabled:
                player_trainer_collective.broadcast(
                    torch.nn.utils.convert_parameters.parameters_to_vector(agent.parameters()), src=1
                )
            return
        sampler = BatchSampler(
            range(len(data[next(iter(data.keys()))])), batch_size=cfg.algo.per_rank_batch_size, drop_last=False
        )

        # Start training
        with timer(
            "Time/train_time", SumMetric, sync_on_compute=cfg.metric.sync_on_compute, process_group=optimization_pg
        ):
            for batch_idxes in sampler:
                train(
                    fabric,
                    agent,
                    actor_optimizer,
                    qf_optimizer,
                    alpha_optimizer,
                    {k: data[k][batch_idxes] for k in data.keys()},
                    aggregator,
                    update,
                    cfg,
                    policy_steps_per_update,
                    group=optimization_pg,
                )
            train_step += group_world_size

        if global_rank == 1:
            player_trainer_collective.broadcast(
                torch.nn.utils.convert_parameters.parameters_to_vector(agent.actor.parameters()), src=1
            )

        if cfg.metric.log_level > 0 and policy_step - last_log >= cfg.metric.log_every:
            # Sync distributed metrics
            metrics = {}
            if aggregator and not aggregator.disabled:
                metrics.update(aggregator.compute())
                aggregator.reset()

            # Sync distributed timers
            if not timer.disabled:
                timers = timer.compute()
                metrics.update({"Time/sps_train": (train_step - last_train) / timers["Time/train_time"]})
                timer.reset()

            if global_rank == 1:
                player_trainer_collective.broadcast_object_list(
                    [metrics], src=1
                )  # Broadcast metrics: fake send with object list between rank-0 and rank-1

            # Reset counters
            last_log = policy_step
            last_train = train_step

        # Checkpoint model on rank-0: send it everything
        if cfg.checkpoint.every > 0 and policy_step - last_checkpoint >= cfg.checkpoint.every:
            last_checkpoint = policy_step
            state = {
                "agent": agent.state_dict(),
                "qf_optimizer": qf_optimizer.state_dict(),
                "actor_optimizer": actor_optimizer.state_dict(),
                "alpha_optimizer": alpha_optimizer.state_dict(),
                "update": update,
                "batch_size": cfg.algo.per_rank_batch_size * (world_collective.world_size - 1),
                "last_log": last_log,
                "last_checkpoint": last_checkpoint,
            }
            ckpt_path = cfg.checkpoint.log_dir + f"/checkpoint/ckpt_{policy_step}_{fabric.global_rank}.ckpt"
            fabric.call(
                "on_checkpoint_trainer",
                fabric=fabric,
                player_trainer_collective=player_trainer_collective,
                ckpt_path=ckpt_path,
                state=state,
            )

        # Update counters
        update += 1
        policy_step += policy_steps_per_update


@register_algorithm(decoupled=True)
def main(fabric: Fabric, cfg: Dict[str, Any]):
    if fabric.world_size == 1:
        raise RuntimeError(
            "Please run the script with the number of devices greater than 1: "
            "`python sheeprl.py exp=sac_decoupled fabric.devices=2 ...`"
        )

    if "minedojo" in cfg.env.wrapper._target_.lower():
        raise ValueError(
            "MineDojo is not currently supported by SAC agent, since it does not take "
            "into consideration the action masks provided by the environment, but needed "
            "in order to play correctly the game. "
            "As an alternative you can use one of the Dreamers' agents."
        )

    world_collective = TorchCollective()
    player_trainer_collective = TorchCollective()
    world_collective.setup(
        backend="nccl" if os.environ.get("LT_ACCELERATOR", None) in ("gpu", "cuda") else "gloo",
        timeout=timedelta(days=1),
    )

    # Create a global group, assigning it to the collective: used by the player to exchange
    # collected experiences with the trainers
    world_collective.create_group(timeout=timedelta(days=1))
    global_rank = world_collective.rank

    # Create a group between rank-0 (player) and rank-1 (trainer), assigning it to the collective:
    # used by rank-1 to send metrics to be tracked by the rank-0 at the end of a training episode
    player_trainer_collective.create_group(ranks=[0, 1], timeout=timedelta(days=1))

    # Create a new group, without assigning it to the collective: in this way the trainers can
    # still communicate with the player through the global group, but they can optimize the agent
    # between themselves
    optimization_pg = world_collective.new_group(
        ranks=list(range(1, world_collective.world_size)), timeout=timedelta(days=1)
    )
    if global_rank == 0:
        player(fabric, cfg, world_collective, player_trainer_collective)
    else:
        trainer(world_collective, player_trainer_collective, optimization_pg, cfg)



from __future__ import annotations

from sheeprl.utils.imports import _IS_SUPER_MARIO_BROS_AVAILABLE

if not _IS_SUPER_MARIO_BROS_AVAILABLE:
    raise ModuleNotFoundError(_IS_SUPER_MARIO_BROS_AVAILABLE)


from typing import Any, Dict, SupportsFloat, Tuple

import gym_super_mario_bros as gsmb
import gymnasium as gym
import numpy as np
from gym_super_mario_bros.actions import COMPLEX_MOVEMENT, RIGHT_ONLY, SIMPLE_MOVEMENT
from gymnasium.core import RenderFrame
from nes_py.wrappers import JoypadSpace

ACTIONS_SPACE_MAP = {"simple": SIMPLE_MOVEMENT, "right_only": RIGHT_ONLY, "complex": COMPLEX_MOVEMENT}


class JoypadSpaceCustomReset(JoypadSpace):
    def reset(self, seed: int | None = None, options: Dict[str, Any] | None = None):
        return self.env.reset(seed=seed, options=options)


class SuperMarioBrosWrapper(gym.Wrapper):
    def __init__(self, id: str, action_space: str = "simple", render_mode: str = "rgb_array"):
        env = gsmb.make(id)
        env = JoypadSpaceCustomReset(env, ACTIONS_SPACE_MAP[action_space])
        super().__init__(env)

        self._render_mode = render_mode
        self.observation_space = gym.spaces.Dict(
            {
                "rgb": gym.spaces.Box(
                    env.observation_space.low,
                    env.observation_space.high,
                    env.observation_space.shape,
                    env.observation_space.dtype,
                )
            }
        )
        self.action_space = gym.spaces.Discrete(env.action_space.n)

    @property
    def render_mode(self) -> str:
        return self._render_mode

    @render_mode.setter
    def render_mode(self, render_mode: str):
        self._render_mode = render_mode

    def step(self, action: np.ndarray | int) -> Tuple[Any, SupportsFloat, bool, bool, Dict[str, Any]]:
        if isinstance(action, np.ndarray):
            action = action.squeeze().item()
        obs, reward, done, info = self.env.step(action)
        converted_obs = {"rgb": obs.copy()}
        return converted_obs, reward, done, False, info

    def render(self) -> RenderFrame | list[RenderFrame] | None:
        rendered_frame: np.ndarray | None = self.env.render(mode=self.render_mode)
        if self.render_mode == "rgb_array" and rendered_frame is not None:
            return rendered_frame.copy()
        return

    def reset(self, *, seed: int | None = None, options: Dict[str, Any] | None = None) -> Tuple[Any, Dict[str, Any]]:
        obs = self.env.reset(seed=seed, options=options)
        converted_obs = {"rgb": obs.copy()}
        return converted_obs, {}



import importlib
import os
import pathlib
import warnings
from pathlib import Path
from typing import Any, Dict

import hydra
import torch
from lightning import Fabric
from lightning.fabric.strategies import STRATEGY_REGISTRY, DDPStrategy, SingleDeviceStrategy, Strategy
from omegaconf import DictConfig, OmegaConf, open_dict

from sheeprl.utils.imports import _IS_MLFLOW_AVAILABLE
from sheeprl.utils.logger import get_logger
from sheeprl.utils.metric import MetricAggregator
from sheeprl.utils.registry import algorithm_registry, evaluation_registry
from sheeprl.utils.timer import timer
from sheeprl.utils.utils import dotdict, print_config


def resume_from_checkpoint(cfg: DictConfig) -> DictConfig:
    ckpt_path = pathlib.Path(cfg.checkpoint.resume_from)
    old_cfg = OmegaConf.load(ckpt_path.parent.parent / "config.yaml")
    old_cfg = dotdict(OmegaConf.to_container(old_cfg, resolve=True, throw_on_missing=True))
    if old_cfg.env.id != cfg.env.id:
        raise ValueError(
            "This experiment is run with a different environment from the one of the experiment you want to restart. "
            f"Got '{cfg.env.id}', but the environment of the experiment of the checkpoint was {old_cfg.env.id}. "
            "Set properly the environment for restarting the experiment."
        )
    if old_cfg.algo.name != cfg.algo.name:
        raise ValueError(
            "This experiment is run with a different algorithm from the one of the experiment you want to restart. "
            f"Got '{cfg.algo.name}', but the algorithm of the experiment of the checkpoint was {old_cfg.algo.name}. "
            "Set properly the algorithm name for restarting the experiment."
        )

    # Remove keys from the `old_cfg` that must not be overridden
    old_cfg.pop("root_dir", None)
    old_cfg.pop("run_name", None)
    old_cfg.checkpoint.pop("resume_from", None)
    # Substitute the config with the old one (except for the parameters removed before)
    # because the experiment must continue with the same parameters
    with open_dict(cfg):
        cfg.merge_with(old_cfg.as_dict())
    return cfg


def run_algorithm(cfg: Dict[str, Any]):
    """Run the algorithm specified in the configuration.

    Args:
        cfg (Dict[str, Any]): the loaded configuration.
    """

    # Torch settings
    os.environ["OMP_NUM_THREADS"] = str(cfg.num_threads)
    torch.set_float32_matmul_precision(cfg.float32_matmul_precision)

    # Given the algorithm's name, retrieve the module where
    # 'cfg.algo.name'.py is contained; from there retrieve the
    # 'register_algorithm'-decorated entrypoint;
    # the entrypoint will be launched by Fabric with 'fabric.launch(entrypoint)'
    module = None
    decoupled = False
    entrypoint = None
    algo_name = cfg.algo.name
    for _module, _algos in algorithm_registry.items():
        for _algo in _algos:
            if algo_name == _algo["name"]:
                module = _module
                entrypoint = _algo["entrypoint"]
                decoupled = _algo["decoupled"]
                break
    if module is None:
        raise RuntimeError(f"Given the algorithm named '{algo_name}', no module has been found to be imported.")
    if entrypoint is None:
        raise RuntimeError(
            f"Given the module and algorithm named '{module}' and '{algo_name}' respectively, "
            "no entrypoint has been found to be imported."
        )
    task = importlib.import_module(f"{module}.{algo_name}")
    utils = importlib.import_module(f"{module}.utils")
    command = task.__dict__[entrypoint]
    kwargs = {}
    if decoupled:
        fabric: Fabric = hydra.utils.instantiate(cfg.fabric, _convert_="all")
        logger = get_logger(fabric, cfg)
        if logger and fabric.is_global_zero:
            fabric._loggers = [logger]
            fabric.logger.log_hyperparams(cfg)
    else:
        strategy = cfg.fabric.get("strategy", "auto")
        if "sac_ae" in module:
            if strategy is not None:
                warnings.warn(
                    "You are running the SAC-AE algorithm you have specified a strategy different than 'ddp': "
                    f"'python sheeprl.py fabric.strategy={strategy}'. This algorithm is run with the "
                    "'lightning.fabric.strategies.DDPStrategy' strategy."
                )
            cfg.fabric.pop("strategy", "auto")
            strategy = DDPStrategy(find_unused_parameters=True)
        elif "finetuning" in algo_name and "p2e" in module:
            # Load exploration configurations
            ckpt_path = pathlib.Path(cfg.checkpoint.exploration_ckpt_path)
            exploration_cfg = OmegaConf.load(ckpt_path.parent.parent / "config.yaml")
            exploration_cfg = dotdict(OmegaConf.to_container(exploration_cfg, resolve=True, throw_on_missing=True))
            if exploration_cfg.env.id != cfg.env.id:
                raise ValueError(
                    "This experiment is run with a different environment from "
                    "the one of the exploration you want to finetune. "
                    f"Got '{cfg.env.id}', but the environment used during exploration was {exploration_cfg.env.id}. "
                    "Set properly the environment for finetuning the experiment."
                )
            kwargs["exploration_cfg"] = exploration_cfg
            # Take environment configs from exploration
            cfg.env.frame_stack = exploration_cfg.env.frame_stack
            cfg.env.screen_size = exploration_cfg.env.screen_size
            cfg.env.action_repeat = exploration_cfg.env.action_repeat
            cfg.env.grayscale = exploration_cfg.env.grayscale
            cfg.env.clip_rewards = exploration_cfg.env.clip_rewards
            cfg.env.frame_stack_dilation = exploration_cfg.env.frame_stack_dilation
            cfg.env.max_episode_steps = exploration_cfg.env.max_episode_steps
            cfg.env.reward_as_observation = exploration_cfg.env.reward_as_observation
            _env_target = cfg.env.wrapper._target_.lower()
            if "minerl" in _env_target or "minedojo" in _env_target:
                cfg.env.max_pitch = exploration_cfg.env.max_pitch
                cfg.env.min_pitch = exploration_cfg.env.min_pitch
                cfg.env.sticky_jump = exploration_cfg.env.sticky_jump
                cfg.env.sticky_attack = exploration_cfg.env.sticky_attack
                cfg.env.break_speed_multiplier = exploration_cfg.env.break_speed_multiplier
            if cfg.buffer.load_from_exploration:
                cfg.fabric.devices = exploration_cfg.fabric.devices
                cfg.fabric.num_nodes = exploration_cfg.fabric.num_nodes
        fabric: Fabric = hydra.utils.instantiate(cfg.fabric, strategy=strategy, _convert_="all")

    if hasattr(cfg, "metric") and cfg.metric is not None:
        predefined_metric_keys = set()
        if not hasattr(utils, "AGGREGATOR_KEYS"):
            warnings.warn(
                f"No 'AGGREGATOR_KEYS' set found for the {algo_name} algorithm under the {module} module. "
                "No metric will be logged.",
                UserWarning,
            )
        else:
            predefined_metric_keys = utils.AGGREGATOR_KEYS
        timer.disabled = cfg.metric.log_level == 0 or cfg.metric.disable_timer
        keys_to_remove = set(cfg.metric.aggregator.metrics.keys()) - predefined_metric_keys
        for k in keys_to_remove:
            cfg.metric.aggregator.metrics.pop(k, None)
        MetricAggregator.disabled = cfg.metric.log_level == 0 or len(cfg.metric.aggregator.metrics) == 0

    # Model Manager
    if hasattr(cfg, "model_manager") and not cfg.model_manager.disabled and cfg.model_manager.models is not None:
        predefined_models_keys = set()
        if not hasattr(utils, "MODELS_TO_REGISTER"):
            warnings.warn(
                f"No 'MODELS_TO_REGISTER' set found for the {algo_name} algorithm under the {module} module. "
                "No model will be registered.",
                UserWarning,
            )
        else:
            predefined_models_keys = utils.MODELS_TO_REGISTER
        keys_to_remove = set(cfg.model_manager.models.keys()) - predefined_models_keys
        for k in keys_to_remove:
            cfg.model_manager.models.pop(k, None)
        cfg.model_manager.disabled == cfg.model_manager.disabled or len(cfg.model_manager.models) == 0
    fabric.launch(command, cfg, **kwargs)


def eval_algorithm(cfg: DictConfig):
    """Run the algorithm specified in the configuration.

    Args:
        cfg (DictConfig): the loaded configuration.
    """
    cfg = dotdict(OmegaConf.to_container(cfg, resolve=True, throw_on_missing=True))

    # Torch settings
    os.environ["OMP_NUM_THREADS"] = str(cfg.num_threads)
    torch.set_float32_matmul_precision(cfg.float32_matmul_precision)

    # TODO: change the number of devices when FSDP will be supported
    accelerator = cfg.fabric.get("accelerator", "auto")
    fabric: Fabric = hydra.utils.instantiate(
        cfg.fabric, accelerator=accelerator, devices=1, num_nodes=1, _convert_="all"
    )

    # Seed everything
    fabric.seed_everything(cfg.seed)

    # Load the checkpoint
    state = fabric.load(cfg.checkpoint_path)

    # Given the algorithm's name, retrieve the module where
    # 'cfg.algo.name'.py is contained; from there retrieve the
    # `register_algorithm`-decorated entrypoint;
    # the entrypoint will be launched by Fabric with `fabric.launch(entrypoint)`
    module = None
    entrypoint = None
    evaluation_file = None
    algo_name = cfg.algo.name
    for _module, _algos in evaluation_registry.items():
        for _algo in _algos:
            if algo_name == _algo["name"]:
                module = _module
                entrypoint = _algo["entrypoint"]
                evaluation_file = _algo["evaluation_file"]
                break
    if module is None:
        raise RuntimeError(f"Given the algorithm named `{algo_name}`, no module has been found to be imported.")
    if entrypoint is None:
        raise RuntimeError(
            f"Given the module and algorithm named `{module}` and `{algo_name}` respectively, "
            "no entrypoint has been found to be imported."
        )
    if evaluation_file is None:
        raise RuntimeError(
            f"Given the module and algorithm named `{module}` and `{algo_name}` respectively, "
            "no evaluation file has been found to be imported."
        )
    task = importlib.import_module(f"{module}.{evaluation_file}")
    command = task.__dict__[entrypoint]
    if getattr(cfg, "disable_grads", True):
        command = torch.no_grad(command)
    fabric.launch(command, cfg, state)


def check_configs(cfg: Dict[str, Any]):
    """Check the validity of the configuration.

    Args:
        cfg (Dict[str, Any]): the loaded configuration to check.
    """
    if cfg.float32_matmul_precision not in {"medium", "high", "highest"}:
        raise ValueError(
            f"Invalid value '{cfg.float32_matmul_precision}' for the 'float32_matmul_precision' parameter. "
            "It must be one of 'medium', 'high' or 'highest'."
        )

    decoupled = False
    algo_name = cfg.algo.name
    for _, _algos in algorithm_registry.items():
        for _algo in _algos:
            if algo_name == _algo["name"]:
                decoupled = _algo["decoupled"]
                break
    strategy = cfg.fabric.strategy
    available_strategies = STRATEGY_REGISTRY.available_strategies()
    if decoupled:
        if isinstance(strategy, str):
            strategy = strategy.lower()
            if not (strategy in available_strategies and "ddp" in strategy):
                raise ValueError(
                    f"{strategy} is currently not supported for decoupled algorithm. "
                    "Please launch the script with a DDP strategy: "
                    "'python sheeprl.py fabric.strategy=ddp'"
                )
        elif (
            "_target_" in strategy
            and issubclass((strategy := hydra.utils.get_class(strategy._target_)), Strategy)
            and not issubclass(strategy, DDPStrategy)
        ):
            raise ValueError(
                f"{strategy.__qualname__} is currently not supported for decoupled algorithms. "
                "Please launch the script with a 'DDP' strategy with 'python sheeprl.py fabric.strategy=ddp'"
            )
    else:
        if isinstance(strategy, str):
            strategy = strategy.lower()
            if strategy != "auto" and not (strategy in available_strategies and "ddp" in strategy):
                warnings.warn(
                    f"Running an algorithm with a strategy ({strategy}) "
                    "different than 'auto' or 'dpp' can cause unexpected problems. "
                    "Please launch the script with a 'DDP' strategy with 'python sheeprl.py fabric.strategy=ddp' "
                    "or the 'auto' one with 'python sheeprl.py fabric.strategy=auto' if you run into any problems.",
                    UserWarning,
                )
        elif (
            "_target_" in strategy
            and issubclass((strategy := hydra.utils.get_class(strategy._target_)), Strategy)
            and not issubclass(strategy, (DDPStrategy, SingleDeviceStrategy))
        ):
            warnings.warn(
                f"Running an algorithm with a strategy ({strategy.__qualname__}) "
                "different than 'SingleDeviceStrategy' or 'DDPStrategy' can cause unexpected problems. "
                "Please launch the script with a 'DDP' strategy with 'python sheeprl.py fabric.strategy=ddp' "
                "or with a single device with 'python sheeprl.py fabric.strategy=auto fabric.devices=1' "
                "if you run into any problems.",
                UserWarning,
            )
    if not (_IS_MLFLOW_AVAILABLE or cfg.model_manager.disabled):
        warnings.warn(
            "MLFlow is not installed. "
            "Please install it with 'pip install mlflow' if you want to use the MLFlow logger and log models. "
            "Setting `cfg.model_manager.disabled=True`",
            UserWarning,
        )
        cfg.model_manager.disabled = True


def check_configs_evaluation(cfg: DictConfig):
    if cfg.float32_matmul_precision not in {"medium", "high", "highest"}:
        raise ValueError(
            f"Invalid value '{cfg.float32_matmul_precision}' for the 'float32_matmul_precision' parameter. "
            "It must be one of 'medium', 'high' or 'highest'."
        )

    if cfg.checkpoint_path is None:
        raise ValueError("You must specify the evaluation checkpoint path")


@hydra.main(version_base="1.3", config_path="configs", config_name="config")
def run(cfg: DictConfig):
    """SheepRL zero-code command line utility."""
    print_config(cfg)
    if cfg.checkpoint.resume_from:
        cfg = resume_from_checkpoint(cfg)
    cfg = dotdict(OmegaConf.to_container(cfg, resolve=True, throw_on_missing=True))
    check_configs(cfg)
    run_algorithm(cfg)


@hydra.main(version_base="1.3", config_path="configs", config_name="eval_config")
def evaluation(cfg: DictConfig):
    # Load the checkpoint configuration
    checkpoint_path = Path(os.path.abspath(cfg.checkpoint_path))
    ckpt_cfg = OmegaConf.load(checkpoint_path.parent.parent / "config.yaml")
    ckpt_cfg.pop("seed", None)

    # Merge the two configs
    with open_dict(cfg):
        capture_video = getattr(cfg.env, "capture_video", True)
        cfg.env = {"capture_video": capture_video, "num_envs": 1}
        cfg.exp = {}
        cfg.algo = {}
        cfg.fabric = {
            "devices": 1,
            "num_nodes": 1,
            "strategy": "auto",
            "accelerator": getattr(cfg.fabric, "accelerator", "auto"),
        }
        cfg.root_dir = str(checkpoint_path.parent.parent.parent.parent)

        # Merge configs
        ckpt_cfg.merge_with(cfg)

        # Update values after merge
        run_name = Path(
            os.path.join(
                os.path.basename(checkpoint_path.parent.parent.parent),
                os.path.basename(checkpoint_path.parent.parent),
                "evaluation",
            )
        )
        ckpt_cfg.run_name = str(run_name)

    # Check the validity of the configuration and run the evaluation
    check_configs_evaluation(ckpt_cfg)
    eval_algorithm(ckpt_cfg)


@hydra.main(version_base="1.3", config_path="configs", config_name="model_manager_config")
def registration(cfg: DictConfig):
    from sheeprl.utils.mlflow import register_model_from_checkpoint

    checkpoint_path = Path(cfg.checkpoint_path)
    ckpt_cfg = OmegaConf.load(checkpoint_path.parent.parent / "config.yaml")

    # Merge the two configs
    with open_dict(cfg):
        cfg.env = ckpt_cfg.env
        cfg.exp_name = ckpt_cfg.exp_name
        cfg.algo = ckpt_cfg.algo
        cfg.distribution = ckpt_cfg.distribution
        cfg.seed = ckpt_cfg.seed

    cfg = dotdict(OmegaConf.to_container(cfg, resolve=True, throw_on_missing=True))
    cfg.to_log = dotdict(OmegaConf.to_container(ckpt_cfg, resolve=True, throw_on_missing=True))

    precision = getattr(ckpt_cfg.fabric, "precision", None)
    fabric = Fabric(devices=1, accelerator="cpu", num_nodes=1, precision=precision)

    # Load the checkpoint
    state = fabric.load(cfg.checkpoint_path)
    # Retrieve the algorithm name, used to import the custom
    # log_models_from_checkpoint function.
    algo_name = cfg.algo.name
    if "decoupled" in cfg.algo.name:
        algo_name = algo_name.replace("_decoupled", "")
    if algo_name.startswith("p2e_dv"):
        algo_name = "_".join(algo_name.split("_")[:2])
    try:
        log_models_from_checkpoint = importlib.import_module(
            f"sheeprl.algos.{algo_name}.utils"
        ).log_models_from_checkpoint
    except Exception as e:
        print(e)
        raise RuntimeError(
            f"Make sure that the algorithm is defined in the `./sheeprl/algos/{algo_name}` folder "
            "and that the `log_models_from_checkpoint` function is defined "
            f"in the `./sheeprl/algos/{algo_name}/utils.py` file."
        )

    fabric.launch(register_model_from_checkpoint, cfg, state, log_models_from_checkpoint)



"""From: https://github.com/toshas/torch_truncnorm/blob/main/TruncatedNormal.py"""

import math
from numbers import Number
from typing import Callable

import torch
import torch.nn.functional as F
from torch import Tensor
from torch.distributions import Bernoulli, Categorical, Distribution, constraints
from torch.distributions.kl import _kl_categorical_categorical, register_kl
from torch.distributions.utils import broadcast_all

from sheeprl.utils.utils import symexp, symlog

CONST_SQRT_2 = math.sqrt(2)
CONST_INV_SQRT_2PI = 1 / math.sqrt(2 * math.pi)
CONST_INV_SQRT_2 = 1 / math.sqrt(2)
CONST_LOG_INV_SQRT_2PI = math.log(CONST_INV_SQRT_2PI)
CONST_LOG_SQRT_2PI_E = 0.5 * math.log(2 * math.pi * math.e)

__all__ = ["OneHotCategoricalValidateArgs", "OneHotCategoricalStraightThroughValidateArgs"]


class TruncatedStandardNormal(Distribution):
    """
    Truncated Standard Normal distribution
    https://people.sc.fsu.edu/~jburkardt/presentations/truncated_normal.pdf
    """

    arg_constraints = {
        "a": constraints.real,
        "b": constraints.real,
    }
    has_rsample = True

    def __init__(self, a, b, validate_args=None):
        self.a, self.b = broadcast_all(a, b)
        if isinstance(a, Number) and isinstance(b, Number):
            batch_shape = torch.Size()
        else:
            batch_shape = self.a.size()
        super(TruncatedStandardNormal, self).__init__(batch_shape, validate_args=validate_args)
        if self.a.dtype != self.b.dtype:
            raise ValueError("Truncation bounds types are different")
        if any((self.a >= self.b).view(-1).tolist()):
            raise ValueError("Incorrect truncation range")
        eps = torch.finfo(self.a.dtype).eps
        self._dtype_min_gt_0 = eps
        self._dtype_max_lt_1 = 1 - eps
        self._little_phi_a = self._little_phi(self.a)
        self._little_phi_b = self._little_phi(self.b)
        self._big_phi_a = self._big_phi(self.a)
        self._big_phi_b = self._big_phi(self.b)
        self._Z = (self._big_phi_b - self._big_phi_a).clamp_min(eps)
        self._log_Z = self._Z.log()
        little_phi_coeff_a = torch.nan_to_num(self.a, nan=math.nan)
        little_phi_coeff_b = torch.nan_to_num(self.b, nan=math.nan)
        self._lpbb_m_lpaa_d_Z = (
            self._little_phi_b * little_phi_coeff_b - self._little_phi_a * little_phi_coeff_a
        ) / self._Z
        self._mean = -(self._little_phi_b - self._little_phi_a) / self._Z
        self._variance = 1 - self._lpbb_m_lpaa_d_Z - ((self._little_phi_b - self._little_phi_a) / self._Z) ** 2
        self._entropy = CONST_LOG_SQRT_2PI_E + self._log_Z - 0.5 * self._lpbb_m_lpaa_d_Z

    @constraints.dependent_property
    def support(self):
        return constraints.interval(self.a, self.b)

    @property
    def mean(self):
        return self._mean

    @property
    def variance(self):
        return self._variance

    @property
    def auc(self):
        return self._Z

    @staticmethod
    def _little_phi(x):
        return (-(x**2) * 0.5).exp() * CONST_INV_SQRT_2PI

    @staticmethod
    def _big_phi(x):
        return 0.5 * (1 + (x * CONST_INV_SQRT_2).erf())

    @staticmethod
    def _inv_big_phi(x):
        return CONST_SQRT_2 * (2 * x - 1).erfinv()

    def cdf(self, value):
        if self._validate_args:
            self._validate_sample(value)
        return ((self._big_phi(value) - self._big_phi_a) / self._Z).clamp(0, 1)

    def icdf(self, value):
        return self._inv_big_phi(self._big_phi_a + value * self._Z)

    def log_prob(self, value):
        if self._validate_args:
            self._validate_sample(value)
        return CONST_LOG_INV_SQRT_2PI - self._log_Z - (value**2) * 0.5

    def rsample(self, sample_shape=torch.Size()):
        shape = self._extended_shape(sample_shape)
        p = torch.empty(shape, device=self.a.device).uniform_(self._dtype_min_gt_0, self._dtype_max_lt_1)
        return self.icdf(p)

    def entropy(self):
        return self._entropy


class TruncatedNormal(TruncatedStandardNormal):
    """
    Truncated Normal distribution
    https://people.sc.fsu.edu/~jburkardt/presentations/truncated_normal.pdf
    """

    has_rsample = True

    def __init__(self, loc, scale, a, b, validate_args=None):
        self.loc, self.scale, a, b = broadcast_all(loc, scale, a, b)
        a = (a - self.loc) / self.scale
        b = (b - self.loc) / self.scale
        super(TruncatedNormal, self).__init__(a, b, validate_args=validate_args)
        self._log_scale = self.scale.log()
        self._mean = self._mean * self.scale + self.loc
        self._variance = self._variance * self.scale**2
        self._entropy += self._log_scale

    def _to_std_rv(self, value):
        return (value - self.loc) / self.scale

    def _from_std_rv(self, value):
        return value * self.scale + self.loc

    def cdf(self, value):
        return super(TruncatedNormal, self).cdf(self._to_std_rv(value))

    def icdf(self, value):
        return self._from_std_rv(super(TruncatedNormal, self).icdf(value))

    def log_prob(self, value):
        return super(TruncatedNormal, self).log_prob(self._to_std_rv(value)) - self._log_scale


# TODO: let distributions be complaint to torch.distributions.Distribution API
# From https://github.com/danijar/dreamerv3/blob/8fa35f83eee1ce7e10f3dee0b766587d0a713a60/dreamerv3/jaxutils.py
class SymlogDistribution:
    def __init__(
        self,
        mode: Tensor,
        dims: int,
        dist: str = "mse",
        agg: str = "sum",
        tol: float = 1e-8,
    ):
        self._mode = mode
        self._dims = tuple([-x for x in range(1, dims + 1)])
        self._dist = dist
        self._agg = agg
        self._tol = tol
        self._batch_shape = mode.shape[: len(mode.shape) - dims]
        self._event_shape = mode.shape[len(mode.shape) - dims :]

    @property
    def mode(self) -> Tensor:
        return symexp(self._mode)

    @property
    def mean(self) -> Tensor:
        return symexp(self._mode)

    def log_prob(self, value: Tensor) -> Tensor:
        assert self._mode.shape == value.shape, (self._mode.shape, value.shape)
        if self._dist == "mse":
            distance = (self._mode - symlog(value)) ** 2
            distance = torch.where(distance < self._tol, 0, distance)
        elif self._dist == "abs":
            distance = torch.abs(self._mode - symlog(value))
            distance = torch.where(distance < self._tol, 0, distance)
        else:
            raise NotImplementedError(self._dist)
        if self._agg == "mean":
            loss = distance.mean(self._dims)
        elif self._agg == "sum":
            loss = distance.sum(self._dims)
        else:
            raise NotImplementedError(self._agg)
        return -loss


class MSEDistribution:
    def __init__(self, mode: Tensor, dims: int, agg: str = "sum"):
        self._mode = mode
        self._dims = tuple([-x for x in range(1, dims + 1)])
        self._agg = agg
        self._batch_shape = mode.shape[: len(mode.shape) - dims]
        self._event_shape = mode.shape[len(mode.shape) - dims :]

    @property
    def mode(self) -> Tensor:
        return self._mode

    @property
    def mean(self) -> Tensor:
        return self._mode

    def log_prob(self, value: Tensor) -> Tensor:
        assert self._mode.shape == value.shape, (self._mode.shape, value.shape)
        distance = (self._mode - value) ** 2
        if self._agg == "mean":
            loss = distance.mean(self._dims)
        elif self._agg == "sum":
            loss = distance.sum(self._dims)
        else:
            raise NotImplementedError(self._agg)
        return -loss


class TwoHotEncodingDistribution:
    def __init__(
        self,
        logits: Tensor,
        dims: int = 0,
        low: int = -20,
        high: int = 20,
        transfwd: Callable[[Tensor], Tensor] = symlog,
        transbwd: Callable[[Tensor], Tensor] = symexp,
    ):
        self.logits = logits
        self.probs = F.softmax(logits, dim=-1)
        self.dims = tuple([-x for x in range(1, dims + 1)])
        self.bins = torch.linspace(low, high, logits.shape[-1], device=logits.device)
        self.low = low
        self.high = high
        self.transfwd = transfwd
        self.transbwd = transbwd
        self._batch_shape = logits.shape[: len(logits.shape) - dims]
        self._event_shape = logits.shape[len(logits.shape) - dims : -1] + (1,)

    @property
    def mean(self) -> Tensor:
        return self.transbwd((self.probs * self.bins).sum(dim=self.dims, keepdim=True))

    @property
    def mode(self) -> Tensor:
        return self.transbwd((self.probs * self.bins).sum(dim=self.dims, keepdim=True))

    def log_prob(self, x: Tensor) -> Tensor:
        x = self.transfwd(x)
        # below in [-1, len(self.bins) - 1]
        below = (self.bins <= x).type(torch.int32).sum(dim=-1, keepdim=True) - 1
        # above in [0, len(self.bins)]
        above = below + 1

        # above in [0, len(self.bins) - 1]
        above = torch.minimum(above, torch.full_like(above, len(self.bins) - 1))
        # below in [0, len(self.bins) - 1]
        below = torch.maximum(below, torch.zeros_like(below))

        equal = below == above
        dist_to_below = torch.where(equal, 1, torch.abs(self.bins[below] - x))
        dist_to_above = torch.where(equal, 1, torch.abs(self.bins[above] - x))
        total = dist_to_below + dist_to_above
        weight_below = dist_to_above / total
        weight_above = dist_to_below / total
        target = (
            F.one_hot(below, len(self.bins)) * weight_below[..., None]
            + F.one_hot(above, len(self.bins)) * weight_above[..., None]
        ).squeeze(-2)
        log_pred = self.logits - torch.logsumexp(self.logits, dim=-1, keepdims=True)
        return (target * log_pred).sum(dim=self.dims)


# Taken from https://github.com/pytorch/pytorch/blob/main/torch/distributions/one_hot_categorical.py
# to add validate args to the categorical distribution
class OneHotCategoricalValidateArgs(Distribution):
    r"""
    Creates a one-hot categorical distribution parameterized by :attr:`probs` or
    :attr:`logits`.

    Samples are one-hot coded vectors of size ``probs.size(-1)``.

    .. note:: The `probs` argument must be non-negative, finite and have a non-zero sum,
              and it will be normalized to sum to 1 along the last dimension. :attr:`probs`
              will return this normalized value.
              The `logits` argument will be interpreted as unnormalized log probabilities
              and can therefore be any real number. It will likewise be normalized so that
              the resulting probabilities sum to 1 along the last dimension. :attr:`logits`
              will return this normalized value.

    See also: :func:`torch.distributions.Categorical` for specifications of
    :attr:`probs` and :attr:`logits`.

    Example::

        >>> # xdoctest: +IGNORE_WANT("non-deterinistic")
        >>> m = OneHotCategoricalValidateArgs(torch.tensor([ 0.25, 0.25, 0.25, 0.25 ]))
        >>> m.sample()  # equal probability of 0, 1, 2, 3
        tensor([ 0.,  0.,  0.,  1.])

    Args:
        probs (Tensor): event probabilities
        logits (Tensor): event log probabilities (unnormalized)
    """
    arg_constraints = {"probs": constraints.simplex, "logits": constraints.real_vector}
    support = constraints.one_hot
    has_enumerate_support = True

    def __init__(self, probs=None, logits=None, validate_args=None):
        self._categorical = Categorical(probs, logits, validate_args=validate_args)
        batch_shape = self._categorical.batch_shape
        event_shape = self._categorical.param_shape[-1:]
        super().__init__(batch_shape, event_shape, validate_args=validate_args)

    def expand(self, batch_shape, _instance=None):
        new = self._get_checked_instance(OneHotCategoricalValidateArgs, _instance)
        batch_shape = torch.Size(batch_shape)
        new._categorical = self._categorical.expand(batch_shape)
        super(OneHotCategoricalValidateArgs, new).__init__(batch_shape, self.event_shape, validate_args=False)
        new._validate_args = self._validate_args
        return new

    def _new(self, *args, **kwargs):
        return self._categorical._new(*args, **kwargs)

    @property
    def _param(self):
        return self._categorical._param

    @property
    def probs(self):
        return self._categorical.probs

    @property
    def logits(self):
        return self._categorical.logits

    @property
    def mean(self):
        return self._categorical.probs

    @property
    def mode(self):
        probs = self._categorical.probs
        mode = probs.argmax(axis=-1)
        return torch.nn.functional.one_hot(mode, num_classes=probs.shape[-1]).to(probs)

    @property
    def variance(self):
        return self._categorical.probs * (1 - self._categorical.probs)

    @property
    def param_shape(self):
        return self._categorical.param_shape

    def sample(self, sample_shape=torch.Size()):
        sample_shape = torch.Size(sample_shape)
        probs = self._categorical.probs
        num_events = self._categorical._num_events
        indices = self._categorical.sample(sample_shape)
        return torch.nn.functional.one_hot(indices, num_events).to(probs)

    def log_prob(self, value):
        if self._validate_args:
            self._validate_sample(value)
        indices = value.max(-1)[1]
        return self._categorical.log_prob(indices)

    def entropy(self):
        return self._categorical.entropy()

    def enumerate_support(self, expand=True):
        n = self.event_shape[0]
        values = torch.eye(n, dtype=self._param.dtype, device=self._param.device)
        values = values.view((n,) + (1,) * len(self.batch_shape) + (n,))
        if expand:
            values = values.expand((n,) + self.batch_shape + (n,))
        return values


class OneHotCategoricalStraightThroughValidateArgs(OneHotCategoricalValidateArgs):
    r"""
    Creates a reparameterizable :class:`OneHotCategoricalValidateArgs` distribution based on the straight-
    through gradient estimator from [1].

    [1] Estimating or Propagating Gradients Through Stochastic Neurons for Conditional Computation
    (Bengio et al, 2013)
    """
    has_rsample = True

    def rsample(self, sample_shape=torch.Size()):
        samples = self.sample(sample_shape)
        probs = self._categorical.probs  # cached via @lazy_property
        return samples + (probs - probs.detach())


@register_kl(OneHotCategoricalValidateArgs, OneHotCategoricalValidateArgs)
def _kl_onehotcategoricalvalidateargs_onehotcategoricalvalidateargs(p, q):
    return _kl_categorical_categorical(p._categorical, q._categorical)


class BernoulliSafeMode(Bernoulli):
    def __init__(self, probs=None, logits=None, validate_args=None):
        super().__init__(probs, logits, validate_args)

    @property
    def mode(self):
        mode = (self.probs > 0.5).to(self.probs)
        return mode



from __future__ import annotations

from typing import TYPE_CHECKING, Any, Dict, Optional, Sequence, Union

import gymnasium as gym
import torch
import torch.nn as nn
from lightning import Fabric
from torch import Tensor
from torch.distributions import Independent

from sheeprl.utils.distribution import OneHotCategoricalStraightThroughValidateArgs
from sheeprl.utils.env import make_env
from sheeprl.utils.imports import _IS_MLFLOW_AVAILABLE
from sheeprl.utils.utils import unwrap_fabric

if TYPE_CHECKING:
    from mlflow.models.model import ModelInfo

    from sheeprl.algos.dreamer_v1.agent import PlayerDV1
    from sheeprl.algos.dreamer_v2.agent import PlayerDV2


AGGREGATOR_KEYS = {
    "Rewards/rew_avg",
    "Game/ep_len_avg",
    "Loss/world_model_loss",
    "Loss/value_loss",
    "Loss/policy_loss",
    "Loss/observation_loss",
    "Loss/reward_loss",
    "Loss/state_loss",
    "Loss/continue_loss",
    "State/post_entropy",
    "State/prior_entropy",
    "State/kl",
    "Params/exploration_amount",
    "Grads/world_model",
    "Grads/actor",
    "Grads/critic",
}
MODELS_TO_REGISTER = {"world_model", "actor", "critic", "target_critic"}


def compute_stochastic_state(logits: Tensor, discrete: int = 32, sample=True, validate_args=False) -> Tensor:
    """
    Compute the stochastic state from the logits computed by the transition or representaiton model.

    Args:
        logits (Tensor): logits from either the representation model or the transition model.
        discrete (int, optional): the size of the Categorical variables.
            Defaults to 32.
        sample (bool): whether or not to sample the stochastic state.
            Default to True.
        validate_args: whether or not to validate distribution arguments.
            Default to False.

    Returns:
        The sampled stochastic state.
    """
    logits = logits.view(*logits.shape[:-1], -1, discrete)
    dist = Independent(OneHotCategoricalStraightThroughValidateArgs(logits=logits, validate_args=validate_args), 1)
    stochastic_state = dist.rsample() if sample else dist.mode
    return stochastic_state


def init_weights(m: nn.Module, mode: str = "normal"):
    """
    Initialize the parameters of the m module acording to the Xavier
    normal method.

    Args:
        m (nn.Module): the module to be initialized.
    """
    if isinstance(m, (nn.Conv2d, nn.ConvTranspose2d, nn.Linear)):
        if mode == "normal":
            nn.init.xavier_normal_(m.weight.data)
        elif mode == "uniform":
            nn.init.xavier_uniform_(m.weight.data)
        elif mode == "zero":
            nn.init.constant_(m.weight.data, 0)
        else:
            raise RuntimeError(f"Unrecognized initialization: {mode}. Choose between: `normal`, `uniform` and `zero`")
        if m.bias is not None:
            nn.init.constant_(m.bias.data, 0)


def compute_lambda_values(
    rewards: Tensor,
    values: Tensor,
    continues: Tensor,
    bootstrap: Optional[Tensor] = None,
    horizon: int = 15,
    lmbda: float = 0.95,
) -> Tensor:
    if bootstrap is None:
        bootstrap = torch.zeros_like(values[-1:])
    agg = bootstrap
    next_val = torch.cat((values[1:], bootstrap), dim=0)
    inputs = rewards + continues * next_val * (1 - lmbda)
    lv = []
    for i in reversed(range(horizon)):
        agg = inputs[i] + continues[i] * lmbda * agg
        lv.append(agg)
    return torch.cat(list(reversed(lv)), dim=0)


@torch.no_grad()
def test(
    player: Union["PlayerDV2", "PlayerDV1"],
    fabric: Fabric,
    cfg: Dict[str, Any],
    log_dir: str,
    test_name: str = "",
    sample_actions: bool = False,
):
    """Test the model on the environment with the frozen model.

    Args:
        player (PlayerDV2 | PlayerDV1): the agent which contains all the models needed to play.
        fabric (Fabric): the fabric instance.
        cfg (Dict[str, Any]): the hyper-parameters.
        log_dir (str): the logging directory.
        test_name (str): the name of the test.
            Default to "".
        sample_actoins (bool): whether or not to sample actions.
            Default to False.
    """
    env: gym.Env = make_env(cfg, cfg.seed, 0, log_dir, "test" + (f"_{test_name}" if test_name != "" else ""))()
    done = False
    cumulative_rew = 0
    device = fabric.device
    next_obs = env.reset(seed=cfg.seed)[0]
    for k in next_obs.keys():
        next_obs[k] = torch.from_numpy(next_obs[k]).view(1, *next_obs[k].shape).float()
    player.num_envs = 1
    player.init_states()
    while not done:
        # Act greedly through the environment
        preprocessed_obs = {}
        for k, v in next_obs.items():
            if k in cfg.algo.cnn_keys.encoder:
                preprocessed_obs[k] = v[None, ...].to(device) / 255 - 0.5
            elif k in cfg.algo.mlp_keys.encoder:
                preprocessed_obs[k] = v[None, ...].to(device)
        real_actions = player.get_greedy_action(
            preprocessed_obs, sample_actions, {k: v for k, v in preprocessed_obs.items() if k.startswith("mask")}
        )
        if player.actor.is_continuous:
            real_actions = torch.cat(real_actions, -1).cpu().numpy()
        else:
            real_actions = torch.cat([real_act.argmax(dim=-1) for real_act in real_actions], dim=-1).cpu().numpy()

        # Single environment step
        next_obs, reward, done, truncated, _ = env.step(real_actions.reshape(env.action_space.shape))
        for k in next_obs.keys():
            next_obs[k] = torch.from_numpy(next_obs[k]).view(1, *next_obs[k].shape).float()
        done = done or truncated or cfg.dry_run
        cumulative_rew += reward
    fabric.print("Test - Reward:", cumulative_rew)
    if cfg.metric.log_level > 0 and len(fabric.loggers) > 0:
        fabric.logger.log_metrics({"Test/cumulative_reward": cumulative_rew}, 0)
    env.close()


def log_models_from_checkpoint(
    fabric: Fabric, env: gym.Env | gym.Wrapper, cfg: Dict[str, Any], state: Dict[str, Any]
) -> Sequence["ModelInfo"]:
    if not _IS_MLFLOW_AVAILABLE:
        raise ModuleNotFoundError(str(_IS_MLFLOW_AVAILABLE))
    import mlflow  # noqa

    from sheeprl.algos.dreamer_v2.agent import build_agent

    # Create the models
    is_continuous = isinstance(env.action_space, gym.spaces.Box)
    is_multidiscrete = isinstance(env.action_space, gym.spaces.MultiDiscrete)
    actions_dim = tuple(
        env.action_space.shape
        if is_continuous
        else (env.action_space.nvec.tolist() if is_multidiscrete else [env.action_space.n])
    )
    world_model, actor, critic, target_critic = build_agent(
        fabric,
        actions_dim,
        is_continuous,
        cfg,
        env.observation_space,
        state["world_model"],
        state["actor"],
        state["critic"],
        state["target_critic"],
    )

    # Log the model, create a new run if `cfg.run_id` is None.
    model_info = {}
    with mlflow.start_run(run_id=cfg.run.id, experiment_id=cfg.experiment.id, run_name=cfg.run.name, nested=True) as _:
        model_info["world_model"] = mlflow.pytorch.log_model(unwrap_fabric(world_model), artifact_path="world_model")
        model_info["actor"] = mlflow.pytorch.log_model(unwrap_fabric(actor), artifact_path="actor")
        model_info["critic"] = mlflow.pytorch.log_model(unwrap_fabric(critic), artifact_path="critic")
        model_info["target_critic"] = mlflow.pytorch.log_model(target_critic, artifact_path="target_critic")
        mlflow.log_dict(cfg.to_log, "config.json")
    return model_info



from __future__ import annotations

from typing import TYPE_CHECKING, Any, Dict, Sequence

import gymnasium as gym
from lightning import Fabric

from sheeprl.algos.droq.agent import build_agent
from sheeprl.algos.sac.utils import AGGREGATOR_KEYS as sac_aggregator_keys
from sheeprl.algos.sac.utils import MODELS_TO_REGISTER as sac_models_to_register
from sheeprl.utils.imports import _IS_MLFLOW_AVAILABLE
from sheeprl.utils.utils import unwrap_fabric

if TYPE_CHECKING:
    from mlflow.models.model import ModelInfo

AGGREGATOR_KEYS = sac_aggregator_keys
MODELS_TO_REGISTER = sac_models_to_register


def log_models_from_checkpoint(
    fabric: Fabric, env: gym.Env | gym.Wrapper, cfg: Dict[str, Any], state: Dict[str, Any]
) -> Sequence["ModelInfo"]:
    if not _IS_MLFLOW_AVAILABLE:
        raise ModuleNotFoundError(str(_IS_MLFLOW_AVAILABLE))
    import mlflow  # noqa

    # Create the models
    agent = build_agent(fabric, cfg, env.observation_space, env.action_space, state["agent"])

    # Log the model, create a new run if `cfg.run_id` is None.
    model_info = {}
    with mlflow.start_run(run_id=cfg.run.id, experiment_id=cfg.experiment.id, run_name=cfg.run.name, nested=True) as _:
        model_info["agent"] = mlflow.pytorch.log_model(unwrap_fabric(agent), artifact_path="agent")
        mlflow.log_dict(cfg.to_log, "config.json")
    return model_info



import platform

from lightning_utilities.core.imports import RequirementCache

_IS_ATARI_AVAILABLE = RequirementCache("gymnasium[atari]")
_IS_ATARI_ROMS_AVAILABLE = RequirementCache("gymnasium[accept-rom-license]")
_IS_BOX2D_AVAILABLE = RequirementCache("gymnasium[box2d]")
_IS_CRAFTER_AVAILABLE = RequirementCache("crafter")
_IS_DIAMBRA_AVAILABLE = RequirementCache("diambra")
_IS_DIAMBRA_ARENA_AVAILABLE = RequirementCache("diambra-arena")
_IS_DMC_AVAILABLE = RequirementCache("dm_control")
_IS_MINEDOJO_AVAILABLE = RequirementCache("minedojo")
_IS_MINERL_0_4_4_AVAILABLE = RequirementCache("minerl==0.4.4")
_IS_MLFLOW_AVAILABLE = RequirementCache("mlflow>=2.8", "mlflow")
_IS_SUPER_MARIO_BROS_AVAILABLE = RequirementCache("gym-super-mario-bros==7.4.0")
_IS_TORCH_GREATER_EQUAL_2_0 = RequirementCache("torch>=2.0")
_IS_WINDOWS = platform.system() == "Windows"



from __future__ import annotations

from sheeprl.utils.imports import _IS_MINERL_0_4_4_AVAILABLE

if not _IS_MINERL_0_4_4_AVAILABLE:
    raise ModuleNotFoundError(_IS_MINERL_0_4_4_AVAILABLE)

import copy
from typing import Any, Dict, Optional, SupportsFloat, Tuple

import gymnasium as gym
import minerl
import numpy as np
from minerl.herobraine.hero import mc

from sheeprl.envs.minerl_envs.navigate import CustomNavigate
from sheeprl.envs.minerl_envs.obtain import CustomObtainDiamond, CustomObtainIronPickaxe

# In order to use the environment as a gym you need to register it with gym
CUSTOM_ENVS = {
    "custom_navigate": CustomNavigate,
    "custom_obtain_diamond": CustomObtainDiamond,
    "custom_obtain_iron_pickaxe": CustomObtainIronPickaxe,
}


N_ALL_ITEMS = len(mc.ALL_ITEMS)
NOOP = {
    "camera": (0, 0),
    "forward": 0,
    "back": 0,
    "left": 0,
    "right": 0,
    "attack": 0,
    "sprint": 0,
    "jump": 0,
    "sneak": 0,
    "craft": "none",
    "nearbyCraft": "none",
    "nearbySmelt": "none",
    "place": "none",
    "equip": "none",
}
ITEM_ID_TO_NAME = dict(enumerate(mc.ALL_ITEMS))
ITEM_NAME_TO_ID = dict(zip(mc.ALL_ITEMS, range(N_ALL_ITEMS)))


class MineRLWrapper(gym.Wrapper):
    """Wrapper for the MineRL environments.

    Args:
        id (str): the id of the environment.
        height (int): the height of the pixels observations.
            Default to 64.
        width (int): the width of the pixels observations.
            Default to 64.
        pitch_limits (Tuple[int, int]): the minimum and maximum angle of the pitch axis.
            Default to (-60, 60).
        seed (int, optional): the seed of the action and observation space.
            Default to None.
        sticky_attack (int, optional): how many times repeat the `attack` action when selected.
            Default to 30.
        sticky_jump (int, optional): how many times repeat the `jump` action when selected.
            Default to 10.
        break_speed_multiplier (int, optional): multiplier to increase block breaking speed.
            Default to 100.
        multihot_inventory (bool): whether to consider all the minecraft elements in the inventory.
            If false, the vector of the inventory will contain only the objects obtainable in the chosen task.
            Default to True.
    """

    def __init__(
        self,
        id: str,
        height: int = 64,
        width: int = 64,
        pitch_limits: Tuple[int, int] = (-60, 60),
        seed: Optional[int] = None,
        sticky_attack: Optional[int] = 30,
        sticky_jump: Optional[int] = 10,
        break_speed_multiplier: Optional[int] = 100,
        multihot_inventory: bool = True,
        **kwargs: Optional[Dict[Any, Any]],
    ):
        self._height = height
        self._width = width
        self._pitch_limits = pitch_limits
        self._sticky_attack = sticky_attack
        self._sticky_jump = sticky_jump
        self._sticky_attack_counter = 0
        self._sticky_jump_counter = 0
        self._break_speed_multiplier = break_speed_multiplier
        self._multihot_inventory = multihot_inventory
        if "navigate" not in id.lower():
            kwargs.pop("extreme", None)

        env = CUSTOM_ENVS[id.lower()](break_speed=break_speed_multiplier, **kwargs).make()
        super().__init__(env)

        # Creation a mapping between the discrete action space and the MineRL action space.
        # The mapping will have the following form:
        #   {
        #       idx_action_0: {<minerl_action_0>}
        #       idx_action_1: {<minerl_action_1>}
        #       ...
        #       idx_action_n: {<minerl_action_n>}
        #   }
        # For instance:
        #   {
        #       0: {}
        #       1: {"attack": 1}
        #       2: {"forward": 1}
        #       ...
        #       n-1: {"craft": "planks"}
        #       n: {"craft": "torch"}
        #   }
        self.ACTIONS_MAP = {0: {}}
        act_idx = 1
        for act in self.env.action_space:
            if isinstance(self.env.action_space[act], minerl.herobraine.hero.spaces.Enum):
                act_val = set(self.env.action_space[act].values.tolist()) - {"none"}
                act_len = len(act_val)
            elif act != "camera":
                act_len = 1
                act_val = [1]
            else:
                act_len = 4
                act_val = [
                    np.array([-15, 0]),
                    np.array([15, 0]),
                    np.array([0, -15]),
                    np.array([0, 15]),
                ]
            action = dict(zip((np.arange(act_len) + act_idx).tolist(), [{act: v} for v in act_val]))
            if act in {"jump", "sneak", "sprint"}:
                action[act_idx]["forward"] = 1
            self.ACTIONS_MAP.update(action)
            act_idx += act_len

        # action and observations space
        self.action_space = gym.spaces.Discrete(len(self.ACTIONS_MAP))

        obs_space = {
            "rgb": gym.spaces.Box(0, 255, (3, 64, 64), np.uint8),
            "life_stats": gym.spaces.Box(0.0, np.array([20.0, 20.0, 300.0]), (3,), np.float32),
            # If multihot_inventory, then all the Minecraft objects are included in the inventory vector.
            # Otherwise, only items that can be included in the inventory with respect to the task are considered
            "inventory": (
                gym.spaces.Box(0.0, np.inf, (N_ALL_ITEMS,), np.float32)
                if multihot_inventory
                else gym.spaces.Box(0.0, np.inf, (len(self.env.observation_space["inventory"]),), np.float32)
            ),
            "max_inventory": (
                gym.spaces.Box(0.0, np.inf, (N_ALL_ITEMS,), np.float32)
                if multihot_inventory
                else gym.spaces.Box(0.0, np.inf, (len(self.env.observation_space["inventory"]),), np.float32)
            ),
        }
        if "compass" in self.env.observation_space.spaces:
            obs_space["compass"] = gym.spaces.Box(-180, 180, (1,), np.float32)
        if "equipped_items" in self.env.observation_space.spaces:
            obs_space["equipment"] = (
                gym.spaces.Box(0.0, 1.0, (N_ALL_ITEMS,), np.int32)
                if multihot_inventory
                else gym.spaces.Box(
                    0.0,
                    1.0,
                    (len(self.env.observation_space["equipped_items"]["mainhand"]["type"].values.tolist()),),
                    np.int32,
                )
            )

        # Mapping from names to ids (index in the vector)
        if not multihot_inventory:
            self.inventory_size = obs_space["inventory"].shape[0]
            self.inventory_item_to_id = dict(zip(self.env.observation_space["inventory"], range(self.inventory_size)))
            if "equipment" in obs_space:
                self.equip_size = obs_space["equipment"].shape[0]
                self.equip_item_to_id = dict(
                    zip(
                        self.env.observation_space["equipped_items"]["mainhand"]["type"].values.tolist(),
                        range(self.equip_size),
                    )
                )
        else:
            self.inventory_item_to_id = ITEM_NAME_TO_ID
            self.inventory_size = N_ALL_ITEMS
            if "equipment" in obs_space:
                self.equip_item_to_id = ITEM_NAME_TO_ID
                self.equip_size = N_ALL_ITEMS
        self.observation_space = gym.spaces.Dict(obs_space)
        self._pos = {
            "pitch": 0.0,
            "yaw": 0.0,
        }
        self._max_inventory = np.zeros(self.inventory_size)
        self._render_mode: str = "rgb_array"
        self.seed(seed=seed)

    @property
    def render_mode(self) -> str | None:
        return self._render_mode

    def __getattr__(self, name):
        return getattr(self.env, name)

    def _convert_actions(self, action: np.ndarray) -> Dict[str, Any]:
        converted_actions = copy.deepcopy(NOOP)
        # Update the NOOP actions to perform the selected action.
        # For instance:
        #   self.ACTIONS_MAP = {
        #       0: {}
        #       1: {"attack": 1}
        #       2: {"forward": 1}
        #       ...
        #       n-1: {"craft": "planks"}
        #       n: {"craft": "torch"}
        #   }
        #   action = np.array([2])
        #   converted_actions = {
        #       "camera": (0, 0),
        #       "forward": 1,
        #       "back": 0,
        #       "left": 0,
        #       "right": 0,
        #       "attack": 0,
        #       "sprint": 0,
        #       "jump": 0,
        #       "sneak": 0,
        #       "craft": "none",
        #       "nearbyCraft": "none",
        #       "nearbySmelt": "none",
        #       "place": "none",
        #       "equip": "none",
        #   }
        converted_actions.update(self.ACTIONS_MAP[action.item()])
        # Add sticky actions
        if self._sticky_attack:
            if converted_actions["attack"]:
                self._sticky_attack_counter = self._sticky_attack
            if self._sticky_attack_counter > 0:
                converted_actions["attack"] = 1
                converted_actions["jump"] = 0
                self._sticky_attack_counter -= 1
        if self._sticky_jump:
            if converted_actions["jump"]:
                self._sticky_jump_counter = self._sticky_jump
            if self._sticky_jump_counter > 0:
                converted_actions["jump"] = 1
                converted_actions["forward"] = 1
                self._sticky_jump_counter -= 1
        return converted_actions

    def _convert_equipment(self, equipment: Dict[str, Any]) -> np.ndarray:
        equip = np.zeros(self.equip_size, dtype=np.int32)
        try:
            equip[self.equip_item_to_id[equipment["mainhand"]["type"]]] = 1
        except KeyError:
            equip[self.equip_item_to_id["air"]] = 1
        return equip

    def _convert_inventory(self, inventory: Dict[str, Any]) -> Dict[str, np.ndarray]:
        # the inventory counts, as a vector with one entry for each Minecraft item
        converted_inventory = {"inventory": np.zeros(self.inventory_size)}
        for i, (item, quantity) in enumerate(inventory.items()):
            # count the items in the inventory
            if item == "air":
                converted_inventory["inventory"][self.inventory_item_to_id[item]] += 1
            else:
                converted_inventory["inventory"][self.inventory_item_to_id[item]] += quantity
        converted_inventory["max_inventory"] = np.maximum(converted_inventory["inventory"], self._max_inventory)
        self._max_inventory = converted_inventory["max_inventory"].copy()
        return converted_inventory

    def _convert_obs(self, obs: Dict[str, Any]) -> Dict[str, np.ndarray]:
        converted_obs = {
            "rgb": obs["pov"].copy().transpose(2, 0, 1),
            "life_stats": np.array(
                [obs["life_stats"]["life"], obs["life_stats"]["food"], obs["life_stats"]["air"]], dtype=np.float32
            ),
            **self._convert_inventory(obs["inventory"]),
        }
        if "equipment" in self.observation_space.spaces:
            converted_obs["equipment"] = self._convert_equipment(obs["equipped_items"])
        if "compass" in self.observation_space.spaces:
            converted_obs["compass"] = obs["compass"]["angle"].reshape(-1)
        return converted_obs

    def seed(self, seed: Optional[int] = None) -> None:
        self.observation_space.seed(seed)
        self.action_space.seed(seed)

    def step(self, actions: np.ndarray) -> Tuple[Dict[str, Any], SupportsFloat, bool, bool, Dict[str, Any]]:
        converted_actions = self._convert_actions(actions)
        next_pitch = self._pos["pitch"] + converted_actions["camera"][0]
        next_yaw = ((self._pos["yaw"] + converted_actions["camera"][1]) + 180) % 360 - 180
        if not (self._pitch_limits[0] <= next_pitch <= self._pitch_limits[1]):
            converted_actions["camera"] = np.array([0, converted_actions["camera"][1]])
            next_pitch = self._pos["pitch"]

        obs, reward, done, info = self.env.step(converted_actions)
        self._pos = {
            "pitch": next_pitch,
            "yaw": next_yaw,
        }
        info = {}
        return self._convert_obs(obs), reward, done, False, info

    def reset(
        self, seed: Optional[int] = None, options: Optional[Dict[str, Any]] = None
    ) -> Tuple[np.ndarray, Dict[str, Any]]:
        obs = self.env.reset()
        self._max_inventory = np.zeros(self.inventory_size)
        self._sticky_attack_counter = 0
        self._sticky_jump_counter = 0
        self._pos = {
            "pitch": 0.0,
            "yaw": 0.0,
        }
        return self._convert_obs(obs), {}

    def render(self, mode: Optional[str] = "rgb_array"):
        return self.env.render(self.render_mode)



import torch
import torch.nn.functional as F
from torch import Tensor


def policy_loss(
    new_logprobs: Tensor,
    logprobs: Tensor,
    advantages: Tensor,
    clip_coef: float,
    reduction: str = "mean",
) -> Tensor:
    """Compute the policy loss for a batch of data, as described in equation (7) of the paper.

        - Compute the difference between the new and old logprobs.
        - Exponentiate it to find the ratio.
        - Use the ratio and advantages to compute the loss as per equation (7).

    Args:
        new_logprobs (Tensor): the log-probs of the new actions.
        logprobs (Tensor): the log-probs of the sampled actions from the environment.
        advantages (Tensor): the advantages.
        clip_coef (float): the clipping coefficient.

    Returns:
        the policy loss
    """
    logratio = new_logprobs - logprobs
    ratio = logratio.exp()

    pg_loss1 = advantages * ratio
    pg_loss2 = advantages * torch.clamp(ratio, 1 - clip_coef, 1 + clip_coef)
    pg_loss = -torch.min(pg_loss1, pg_loss2)
    reduction = reduction.lower()
    if reduction == "none":
        return pg_loss
    elif reduction == "mean":
        return pg_loss.mean()
    elif reduction == "sum":
        return pg_loss.sum()
    else:
        raise ValueError(f"Unrecognized reduction: {reduction}")


def value_loss(
    new_values: Tensor,
    old_values: Tensor,
    returns: Tensor,
    clip_coef: float,
    clip_vloss: bool,
    reduction: str = "mean",
) -> Tensor:
    if not clip_vloss:
        values_pred = new_values
        # return F.mse_loss(values_pred, returns, reduction=reduction)
    else:
        values_pred = old_values + torch.clamp(new_values - old_values, -clip_coef, clip_coef)
        # return torch.max((new_values - returns) ** 2, (values_pred - returns) ** 2).mean()
    return F.mse_loss(values_pred, returns, reduction=reduction)


def entropy_loss(entropy: Tensor, reduction: str = "mean") -> Tensor:
    ent_loss = -entropy
    reduction = reduction.lower()
    if reduction == "none":
        return ent_loss
    elif reduction == "mean":
        return ent_loss.mean()
    elif reduction == "sum":
        return ent_loss.sum()
    else:
        raise ValueError(f"Unrecognized reduction: {reduction}")



from __future__ import annotations

from typing import TYPE_CHECKING, Any, Dict, Sequence

import gymnasium as gym
import torch
from lightning import Fabric

from sheeprl.algos.ppo.utils import AGGREGATOR_KEYS as ppo_aggregator_keys
from sheeprl.algos.ppo.utils import MODELS_TO_REGISTER as ppo_models_to_register
from sheeprl.algos.ppo_recurrent.agent import RecurrentPPOAgent, build_agent
from sheeprl.utils.env import make_env
from sheeprl.utils.imports import _IS_MLFLOW_AVAILABLE
from sheeprl.utils.utils import unwrap_fabric

if TYPE_CHECKING:
    from mlflow.models.model import ModelInfo


AGGREGATOR_KEYS = ppo_aggregator_keys
MODELS_TO_REGISTER = ppo_models_to_register


@torch.no_grad()
def test(agent: "RecurrentPPOAgent", fabric: Fabric, cfg: Dict[str, Any], log_dir: str):
    env = make_env(cfg, None, 0, log_dir, "test", vector_env_idx=0)()
    agent.eval()
    done = False
    cumulative_rew = 0
    agent.num_envs = 1
    with fabric.device:
        o = env.reset(seed=cfg.seed)[0]
        next_obs = {
            k: torch.as_tensor(o[k], dtype=torch.float32, device=fabric.device).view(1, 1, -1, *o[k].shape[-2:]) / 255
            for k in cfg.algo.cnn_keys.encoder
        }
        next_obs.update(
            {
                k: torch.as_tensor(o[k], dtype=torch.float32, device=fabric.device).view(1, 1, -1)
                for k in cfg.algo.mlp_keys.encoder
            }
        )
        state = (
            torch.zeros(1, 1, agent.rnn_hidden_size, device=fabric.device),
            torch.zeros(1, 1, agent.rnn_hidden_size, device=fabric.device),
        )
        actions = torch.zeros(1, 1, sum(agent.actions_dim), device=fabric.device)
    while not done:
        # Act greedly through the environment
        actions, state = agent.get_greedy_actions(next_obs, state, actions)
        if agent.is_continuous:
            real_actions = torch.cat(actions, -1)
            actions = torch.cat(actions, dim=-1).view(1, 1, -1)
        else:
            real_actions = torch.cat([act.argmax(dim=-1) for act in actions], dim=-1)
            actions = torch.cat([act for act in actions], dim=-1).view(1, 1, -1)

        # Single environment step
        o, reward, done, truncated, info = env.step(real_actions.cpu().numpy().reshape(env.action_space.shape))
        done = done or truncated
        cumulative_rew += reward
        with fabric.device:
            next_obs = {
                k: torch.as_tensor(o[k], dtype=torch.float32).view(1, 1, -1, *o[k].shape[-2:]) / 255
                for k in cfg.algo.cnn_keys.encoder
            }
            next_obs.update(
                {k: torch.as_tensor(o[k], dtype=torch.float32).view(1, 1, -1) for k in cfg.algo.mlp_keys.encoder}
            )

        if cfg.dry_run:
            done = True
    fabric.print("Test - Reward:", cumulative_rew)
    if cfg.metric.log_level > 0:
        fabric.log_dict({"Test/cumulative_reward": cumulative_rew}, 0)
    env.close()


def log_models_from_checkpoint(
    fabric: Fabric, env: gym.Env | gym.Wrapper, cfg: Dict[str, Any], state: Dict[str, Any]
) -> Sequence["ModelInfo"]:
    if not _IS_MLFLOW_AVAILABLE:
        raise ModuleNotFoundError(str(_IS_MLFLOW_AVAILABLE))
    import mlflow  # noqa

    # Create the models
    is_continuous = isinstance(env.action_space, gym.spaces.Box)
    is_multidiscrete = isinstance(env.action_space, gym.spaces.MultiDiscrete)
    actions_dim = tuple(
        env.action_space.shape
        if is_continuous
        else (env.action_space.nvec.tolist() if is_multidiscrete else [env.action_space.n])
    )
    agent = build_agent(fabric, actions_dim, is_continuous, cfg, env.observation_space, state["agent"])

    # Log the model, create a new run if `cfg.run_id` is None.
    model_info = {}
    with mlflow.start_run(run_id=cfg.run.id, experiment_id=cfg.experiment.id, run_name=cfg.run.name, nested=True) as _:
        model_info["agent"] = mlflow.pytorch.log_model(unwrap_fabric(agent), artifact_path="agent")
        mlflow.log_dict(cfg.to_log, "config.json")
    return model_info



from rich.console import Console
from rich.table import Table

from sheeprl.utils.registry import algorithm_registry, evaluation_registry


def available_agents():
    table = Table(title="SheepRL Agents")
    table.add_column("Module")
    table.add_column("Algorithm")
    table.add_column("Entrypoint")
    table.add_column("Decoupled")
    table.add_column("Evaluated by")

    for module, implementations in algorithm_registry.items():
        for algo in implementations:
            evaluated_by = "Undefined"
            if module in evaluation_registry:
                for evaluation in evaluation_registry[module]:
                    if algo["name"] == evaluation["name"]:
                        evaluation_file = evaluation["evaluation_file"]
                        evaluation_entrypoint = evaluation["entrypoint"]
                        evaluated_by = module + "." + evaluation_file + "." + evaluation_entrypoint
                        break
            table.add_row(
                module,
                algo["name"],
                algo["entrypoint"],
                str(algo["decoupled"]),
                evaluated_by,
            )

    console = Console()
    console.print(table)


if __name__ == "__main__":
    available_agents()



import copy
from math import prod
from typing import Any, Dict, Optional, Sequence, SupportsFloat, Tuple, Union

import gymnasium
import torch
import torch.nn as nn
from lightning import Fabric
from lightning.fabric.wrappers import _FabricModule
from numpy.typing import NDArray
from torch import Tensor

from sheeprl.models.models import MLP

LOG_STD_MAX = 2
LOG_STD_MIN = -5


class SACCritic(nn.Module):
    def __init__(self, observation_dim: int, hidden_size: int = 256, num_critics: int = 1):
        """The SAC critic. The architecture is the one specified in https://arxiv.org/abs/1812.05905

        Args:
            observation_dim (int): the input dimensions. Can be either an integer
                or a sequence of integers.
            hidden_size (int): the hidden sizes for both of the two-layer MLP.
                Defaults to 256.
            num_critics (int, optional): the number of critic values to output.
                This is useful if one wants to have a single shared backbone that outputs
                `num_critics` critic values.
                Defaults to 1.
        """
        super().__init__()
        self.model = MLP(
            input_dims=observation_dim,
            output_dim=num_critics,
            hidden_sizes=(hidden_size, hidden_size),
            activation=nn.ReLU,
            flatten_dim=None,
        )

    def forward(self, obs: Tensor, action: Tensor) -> Tensor:
        """Return the Q-value conditioned on the observation and the action

        Args:
            obs (Tensor): input observation
            action (Tensor): input action

        Returns:
            q-value
        """
        x = torch.cat([obs, action], -1)
        return self.model(x)


class SACActor(nn.Module):
    def __init__(
        self,
        observation_dim: int,
        action_dim: int,
        distribution_cfg: Dict[str, Any],
        hidden_size: int = 256,
        action_low: Union[SupportsFloat, NDArray] = -1.0,
        action_high: Union[SupportsFloat, NDArray] = 1.0,
    ):
        """The SAC critic. The architecture is the one specified in https://arxiv.org/abs/1812.05905

        Args:
            observation_dim (int): the input dimensions. Can be either an integer
                or a sequence of integers.
            action_dim (int): the action dimension.
            distribution_cfg (Dict[str, Any]): the configs of the distributions.
            hidden_size (int): the hidden sizes for both of the two-layer MLP.
                Defaults to 256.
            action_low (Union[SupportsFloat, NDArray], optional): the action lower bound.
                Defaults to -1.0.
            action_high (Union[SupportsFloat, NDArray], optional): the action higher bound.
                Defaults to 1.0.
        """
        super().__init__()
        self.distribution_cfg = distribution_cfg

        self.model = MLP(input_dims=observation_dim, hidden_sizes=(hidden_size, hidden_size), flatten_dim=None)
        self.fc_mean = nn.Linear(self.model.output_dim, action_dim)
        self.fc_logstd = nn.Linear(self.model.output_dim, action_dim)

        # Action rescaling buffers
        self.register_buffer("action_scale", torch.tensor((action_high - action_low) / 2.0, dtype=torch.float32))
        self.register_buffer("action_bias", torch.tensor((action_high + action_low) / 2.0, dtype=torch.float32))

    def forward(self, obs: Tensor) -> Tuple[Tensor, Tensor]:
        """Given an observation, it returns a tanh-squashed
        sampled action (correctly rescaled to the environment action bounds) and its
        log-prob (as defined in Eq. 26 of https://arxiv.org/abs/1812.05905)

        Args:
            obs (Tensor): the observation tensor

        Returns:
            tanh-squashed action, rescaled to the environment action bounds
            action log-prob
        """
        x = self.model(obs)
        mean = self.fc_mean(x)
        log_std = self.fc_logstd(x)
        std = torch.clamp(log_std, LOG_STD_MIN, LOG_STD_MAX).exp()
        return self.get_actions_and_log_probs(mean, std)

    def get_actions_and_log_probs(self, mean: Tensor, std: Tensor):
        """Given the mean and the std of a Normal distribution, it returns a tanh-squashed
        sampled action (correctly rescaled to the environment action bounds) and its
        log-prob (as defined in Eq. 26 of https://arxiv.org/abs/1812.05905)

        Args:
            mean (Tensor): the mean of the distribution
            std (Tensor): the standard deviation of the distribution

        Returns:
            tanh-squashed action, rescaled to the environment action bounds
            action log-prob
        """
        normal = torch.distributions.Normal(mean, std, validate_args=self.distribution_cfg.validate_args)

        # Reparameterization trick (mean + std * N(0,1))
        x_t = normal.rsample()

        # Squash sample
        y_t = torch.tanh(x_t)

        # Action sampled from a Tanh transformed Gaussian distribution
        action = y_t * self.action_scale + self.action_bias

        # Change of variable for probability distributions
        # Eq. 26 of https://arxiv.org/abs/1812.05905
        log_prob = normal.log_prob(x_t)
        log_prob -= torch.log(self.action_scale * (1 - y_t.pow(2)) + 1e-6)

        # Log-prob of independent actions is the sum of the log-probs
        log_prob = log_prob.sum(-1, keepdim=True)

        return action, log_prob

    def get_greedy_actions(self, obs: Tensor) -> Tensor:
        """Get the action given the input observation greedily

        Args:
            obs (Tensor): input observation

        Returns:
            action
        """
        x = self.model(obs)
        mean = self.fc_mean(x)
        mean = torch.tanh(mean) * self.action_scale + self.action_bias
        return mean


class SACAgent(nn.Module):
    def __init__(
        self,
        actor: Union[SACActor, _FabricModule],
        critics: Sequence[Union[SACCritic, _FabricModule]],
        target_entropy: float,
        alpha: float = 1.0,
        tau: float = 0.005,
        device: torch.device = torch.device("cpu"),
    ) -> None:
        super().__init__()

        # Actor and critics
        self._num_critics = len(critics)
        self.actor = actor
        self.critics = critics

        # Automatic entropy tuning
        self._target_entropy = torch.tensor(target_entropy, device=device)
        self._log_alpha = torch.nn.Parameter(torch.log(torch.tensor([alpha], device=device)), requires_grad=True)

        # EMA tau
        self._tau = tau

    def __setattr__(self, name: str, value: Union[Tensor, nn.Module]) -> None:
        # Taken from https://github.com/pytorch/pytorch/pull/92044
        # Check if a property setter exists. If it does, use it.
        class_attr = getattr(self.__class__, name, None)
        if isinstance(class_attr, property) and class_attr.fset is not None:
            return class_attr.fset(self, value)
        super().__setattr__(name, value)

    @property
    def critics(self) -> nn.ModuleList:
        return self.qfs

    @critics.setter
    def critics(self, critics: Sequence[Union[SACCritic, _FabricModule]]) -> None:
        self._qfs = nn.ModuleList(critics)

        # Create target critic unwrapping the DDP module from the critics to prevent
        # `RuntimeError: DDP Pickling/Unpickling are only supported when using DDP with the default process group.
        # That is, when you have called init_process_group and have not passed process_group
        # argument to DDP constructor`.
        # This happens when we're using the decoupled version of SAC for example
        qfs_unwrapped_modules = []
        for critic in critics:
            if hasattr(critic, "module"):
                critic_module = critic.module
            else:
                critic_module = critic
            qfs_unwrapped_modules.append(critic_module)
        self._qfs_unwrapped = nn.ModuleList(qfs_unwrapped_modules)
        self._qfs_target = copy.deepcopy(self._qfs_unwrapped)
        for p in self._qfs_target.parameters():
            p.requires_grad = False
        return

    @property
    def num_critics(self) -> int:
        return self._num_critics

    @property
    def qfs(self) -> nn.ModuleList:
        return self._qfs

    @property
    def qfs_unwrapped(self) -> nn.ModuleList:
        return self._qfs_unwrapped

    @property
    def actor(self) -> Union[SACActor, _FabricModule]:
        return self._actor

    @actor.setter
    def actor(self, actor: Union[SACActor, _FabricModule]) -> None:
        self._actor = actor
        return

    @property
    def qfs_target(self) -> nn.ModuleList:
        return self._qfs_target

    @property
    def alpha(self) -> float:
        return self._log_alpha.exp().item()

    @property
    def target_entropy(self) -> Tensor:
        return self._target_entropy

    @property
    def log_alpha(self) -> Tensor:
        return self._log_alpha

    def get_actions_and_log_probs(self, obs: Tensor) -> Tuple[Tensor, Tensor]:
        return self.actor(obs)

    def get_greedy_actions(self, obs: Tensor) -> Tensor:
        return self.actor.get_greedy_actions(obs)

    def get_q_values(self, obs: Tensor, action: Tensor) -> Tensor:
        return torch.cat([self.qfs[i](obs, action) for i in range(len(self.qfs))], dim=-1)

    @torch.no_grad()
    def get_target_q_values(self, obs: Tensor, action: Tensor) -> Tensor:
        return torch.cat([self.qfs_target[i](obs, action) for i in range(len(self.qfs))], dim=-1)

    @torch.no_grad()
    def get_next_target_q_values(self, next_obs: Tensor, rewards: Tensor, dones: Tensor, gamma: float):
        # Get q-values for the next observations and actions, estimated by the target q-functions
        next_state_actions, next_state_log_pi = self.get_actions_and_log_probs(next_obs)
        qf_next_target = self.get_target_q_values(next_obs, next_state_actions)
        min_qf_next_target = torch.min(qf_next_target, dim=-1, keepdim=True)[0] - self.alpha * next_state_log_pi
        next_qf_value = rewards + (1 - dones) * gamma * min_qf_next_target
        return next_qf_value

    @torch.no_grad()
    def qfs_target_ema(self) -> None:
        for param, target_param in zip(self.qfs_unwrapped.parameters(), self.qfs_target.parameters()):
            target_param.data.copy_(self._tau * param.data + (1 - self._tau) * target_param.data)


def build_agent(
    fabric: Fabric,
    cfg: Dict[str, Any],
    obs_space: gymnasium.spaces.Dict,
    action_space: gymnasium.spaces.Box,
    agent_state: Optional[Dict[str, Tensor]] = None,
) -> SACAgent:
    act_dim = prod(action_space.shape)
    obs_dim = sum([prod(obs_space[k].shape) for k in cfg.algo.mlp_keys.encoder])
    actor = SACActor(
        observation_dim=obs_dim,
        action_dim=act_dim,
        distribution_cfg=cfg.distribution,
        hidden_size=cfg.algo.actor.hidden_size,
        action_low=action_space.low,
        action_high=action_space.high,
    )
    critics = [
        SACCritic(observation_dim=obs_dim + act_dim, hidden_size=cfg.algo.critic.hidden_size, num_critics=1)
        for _ in range(cfg.algo.critic.n)
    ]
    target_entropy = -act_dim
    agent = SACAgent(actor, critics, target_entropy, alpha=cfg.algo.alpha.alpha, tau=cfg.algo.tau, device=fabric.device)
    if agent_state:
        agent.load_state_dict(agent_state)
    agent.actor = fabric.setup_module(agent.actor)
    agent.critics = [fabric.setup_module(critic) for critic in agent.critics]

    return agent



from __future__ import annotations

import warnings
from typing import TYPE_CHECKING, Any, Dict, Sequence

import gymnasium as gym
import torch
from lightning import Fabric
from lightning.fabric.wrappers import _FabricModule

from sheeprl.algos.sac.agent import SACActor, build_agent
from sheeprl.utils.env import make_env
from sheeprl.utils.imports import _IS_MLFLOW_AVAILABLE
from sheeprl.utils.utils import unwrap_fabric

if TYPE_CHECKING:
    from mlflow.models.model import ModelInfo

AGGREGATOR_KEYS = {
    "Rewards/rew_avg",
    "Game/ep_len_avg",
    "Loss/value_loss",
    "Loss/policy_loss",
    "Loss/alpha_loss",
}
MODELS_TO_REGISTER = {"agent"}


@torch.no_grad()
def test(actor: SACActor, fabric: Fabric, cfg: Dict[str, Any], log_dir: str):
    env = make_env(cfg, None, 0, log_dir, "test", vector_env_idx=0)()
    actor.eval()
    done = False
    cumulative_rew = 0
    with fabric.device:
        o = env.reset(seed=cfg.seed)[0]
        next_obs = torch.cat(
            [torch.as_tensor(o[k], dtype=torch.float32) for k in cfg.algo.mlp_keys.encoder], dim=-1
        ).unsqueeze(
            0
        )  # [N_envs, N_obs]
    while not done:
        # Act greedly through the environment
        action = actor.get_greedy_actions(next_obs)

        # Single environment step
        next_obs, reward, done, truncated, info = env.step(action.cpu().numpy().reshape(env.action_space.shape))
        done = done or truncated
        cumulative_rew += reward
        with fabric.device:
            next_obs = torch.cat(
                [torch.as_tensor(next_obs[k], dtype=torch.float32) for k in cfg.algo.mlp_keys.encoder], dim=-1
            )

        if cfg.dry_run:
            done = True
    fabric.print("Test - Reward:", cumulative_rew)
    if cfg.metric.log_level > 0:
        fabric.logger.log_metrics({"Test/cumulative_reward": cumulative_rew}, 0)
    env.close()


def log_models(
    cfg: Dict[str, Any],
    models_to_log: Dict[str, torch.nn.Module | _FabricModule],
    run_id: str,
    experiment_id: str | None = None,
    run_name: str | None = None,
) -> Dict[str, "ModelInfo"]:
    if not _IS_MLFLOW_AVAILABLE:
        raise ModuleNotFoundError(str(_IS_MLFLOW_AVAILABLE))
    import mlflow  # noqa

    with mlflow.start_run(run_id=run_id, experiment_id=experiment_id, run_name=run_name, nested=True) as _:
        model_info = {}
        unwrapped_models = {}
        for k in cfg.model_manager.models.keys():
            if k not in models_to_log:
                warnings.warn(f"Model {k} not found in models_to_log, skipping.", category=UserWarning)
                continue
            unwrapped_models[k] = unwrap_fabric(models_to_log[k])
            model_info[k] = mlflow.pytorch.log_model(unwrapped_models[k], artifact_path=k)
        mlflow.log_dict(cfg, "config.json")
    return model_info


def log_models_from_checkpoint(
    fabric: Fabric, env: gym.Env | gym.Wrapper, cfg: Dict[str, Any], state: Dict[str, Any]
) -> Sequence["ModelInfo"]:
    if not _IS_MLFLOW_AVAILABLE:
        raise ModuleNotFoundError(str(_IS_MLFLOW_AVAILABLE))
    import mlflow  # noqa

    # Create the models
    agent = build_agent(fabric, cfg, env.observation_space, env.action_space, state["agent"])

    # Log the model, create a new run if `cfg.run_id` is None.
    model_info = {}
    with mlflow.start_run(run_id=cfg.run.id, experiment_id=cfg.experiment.id, run_name=cfg.run.name, nested=True) as _:
        model_info["agent"] = mlflow.pytorch.log_model(unwrap_fabric(agent), artifact_path="agent")
        mlflow.log_dict(cfg.to_log, "config.json")
    return model_info



from __future__ import annotations

from typing import Any, Dict

import gymnasium as gym
from lightning import Fabric

from sheeprl.algos.sac.agent import build_agent
from sheeprl.algos.sac.utils import test
from sheeprl.utils.env import make_env
from sheeprl.utils.logger import get_log_dir, get_logger
from sheeprl.utils.registry import register_evaluation


@register_evaluation(algorithms=["sac", "sac_decoupled"])
def evaluate(fabric: Fabric, cfg: Dict[str, Any], state: Dict[str, Any]):
    logger = get_logger(fabric, cfg)
    if logger and fabric.is_global_zero:
        fabric._loggers = [logger]
        fabric.logger.log_hyperparams(cfg)
    log_dir = get_log_dir(fabric, cfg.root_dir, cfg.run_name)
    fabric.print(f"Log dir: {log_dir}")

    env = make_env(
        cfg,
        cfg.seed,
        0,
        log_dir,
        "test",
        vector_env_idx=0,
    )()
    action_space = env.action_space
    observation_space = env.observation_space
    if not isinstance(action_space, gym.spaces.Box):
        raise ValueError("Only continuous action space is supported for the SAC agent")
    if not isinstance(observation_space, gym.spaces.Dict):
        raise RuntimeError(f"Unexpected observation type, should be of type Dict, got: {observation_space}")
    if len(cfg.algo.mlp_keys.encoder) == 0:
        raise RuntimeError("You should specify at least one MLP key for the encoder: `mlp_keys.encoder=[state]`")
    for k in cfg.algo.mlp_keys.encoder:
        if len(observation_space[k].shape) > 1:
            raise ValueError(
                "Only environments with vector-only observations are supported by the SAC agent. "
                f"Provided environment: {cfg.env.id}"
            )
    fabric.print("Encoder MLP keys:", cfg.algo.mlp_keys.encoder)

    agent = build_agent(fabric, cfg, observation_space, action_space, state["agent"])
    test(agent.actor, fabric, cfg, log_dir)



from __future__ import annotations

from sheeprl.utils.imports import _IS_CRAFTER_AVAILABLE

if not _IS_CRAFTER_AVAILABLE:
    raise ModuleNotFoundError(_IS_CRAFTER_AVAILABLE)

from typing import Any, Dict, List, Optional, Sequence, SupportsFloat, Tuple, Union

import crafter
import gymnasium as gym
import numpy as np
from gymnasium import spaces
from gymnasium.core import RenderFrame


class CrafterWrapper(gym.Wrapper):
    def __init__(self, id: str, screen_size: Sequence[int, int] | int, seed: int | None = None) -> None:
        assert id in {"crafter_reward", "crafter_nonreward"}
        if isinstance(screen_size, int):
            screen_size = (screen_size,) * 2

        env = crafter.Env(size=screen_size, seed=seed, reward=(id == "crafter_reward"))
        super().__init__(env)
        self.observation_space = spaces.Dict(
            {
                "rgb": spaces.Box(
                    self.env.observation_space.low,
                    self.env.observation_space.high,
                    self.env.observation_space.shape,
                    self.env.observation_space.dtype,
                )
            }
        )
        self.action_space = spaces.Discrete(self.env.action_space.n)
        self.reward_range = self.env.reward_range or (-np.inf, np.inf)
        self.observation_space.seed(seed)
        self.action_space.seed(seed)
        # render
        self._render_mode: str = "rgb_array"
        # metadata
        self._metadata = {"render_fps": 30}

    @property
    def render_mode(self) -> str | None:
        return self._render_mode

    def _convert_obs(self, obs: np.ndarray) -> Dict[str, np.ndarray]:
        return {"rgb": obs}

    def step(self, action: Any) -> Tuple[Any, SupportsFloat, bool, bool, Dict[str, Any]]:
        obs, reward, done, info = self.env.step(action)
        return self._convert_obs(obs), reward, done, False, info

    def reset(
        self, *, seed: Optional[int] = None, options: Optional[Dict[str, Any]] = None
    ) -> Tuple[Any, Dict[str, Any]]:
        self.env._seed = seed
        obs = self.env.reset()
        return self._convert_obs(obs), {}

    def render(self) -> Optional[Union[RenderFrame, List[RenderFrame]]]:
        return self.env.render()

    def close(self) -> None:
        return



from __future__ import annotations

import os
import pathlib
from typing import Any, Dict, Optional, Sequence, Union

from lightning.fabric import Fabric
from lightning.fabric.plugins.collectives import TorchCollective
from torch import Tensor

from sheeprl.data.buffers import EnvIndependentReplayBuffer, EpisodeBuffer, ReplayBuffer


class CheckpointCallback:
    """Callback to checkpoint the training.
    Three methods are defined to checkpoint the models, the optimizers, and the replay buffers during the training:
        1. `on_checkpoint_coupled`: The method called by all processes in coupled algorithms,
            the process on rank-0 gets the buffers from all the processes and saves the state of the training.
        2. `on_checkpoint_player`: called by the player process of decoupled algorithms (rank-0),
            it receives the state from the trainer of rank-1 and, if required, adds the replay_buffer to the state.
        3. `on_checkpoint_trainer`: called by the rank-1 trainer process of decoupled algorithms that
            sends the state to the player process (rank-0).

    When the buffer is added to the state of the checkpoint, it is assumed that the episode is truncated.
    """

    def __init__(self, keep_last: int | None = None) -> None:
        self.keep_last = keep_last

    def on_checkpoint_coupled(
        self,
        fabric: Fabric,
        ckpt_path: str,
        state: Dict[str, Any],
        replay_buffer: Optional[Union["EnvIndependentReplayBuffer", "ReplayBuffer", "EpisodeBuffer"]] = None,
    ):
        if replay_buffer is not None:
            rb_state = self._ckpt_rb(replay_buffer)
            state["rb"] = replay_buffer
            if fabric.world_size > 1:
                # We need to collect the buffers from all the ranks
                # The collective it is needed because the `gather_object` function is not implemented in Fabric
                checkpoint_collective = TorchCollective()
                # gloo is the torch.distributed backend that works on cpu
                checkpoint_collective.create_group(backend="gloo", ranks=list(range(fabric.world_size)))
                gathered_rb = [None for _ in range(fabric.world_size)]
                if fabric.global_rank == 0:
                    checkpoint_collective.gather_object(replay_buffer, gathered_rb)
                    state["rb"] = gathered_rb
                else:
                    checkpoint_collective.gather_object(replay_buffer, None)
        fabric.save(ckpt_path, state)
        if replay_buffer is not None:
            self._experiment_consistent_rb(replay_buffer, rb_state)
        if fabric.is_global_zero and self.keep_last:
            self._delete_old_checkpoints(pathlib.Path(ckpt_path).parent)

    def on_checkpoint_player(
        self,
        fabric: Fabric,
        player_trainer_collective: TorchCollective,
        ckpt_path: str,
        replay_buffer: Optional["ReplayBuffer"] = None,
    ):
        state = [None]
        player_trainer_collective.broadcast_object_list(state, src=1)
        state = state[0]
        if replay_buffer is not None:
            rb_state = self._ckpt_rb(replay_buffer)
            state["rb"] = replay_buffer
        fabric.save(ckpt_path, state)
        if replay_buffer is not None:
            self._experiment_consistent_rb(replay_buffer, rb_state)
        if fabric.is_global_zero and self.keep_last:
            self._delete_old_checkpoints(pathlib.Path(ckpt_path).parent)

    def on_checkpoint_trainer(
        self, fabric: Fabric, player_trainer_collective: TorchCollective, state: Dict[str, Any], ckpt_path: str
    ):
        if fabric.global_rank == 1:
            player_trainer_collective.broadcast_object_list([state], src=1)
        fabric.save(ckpt_path, state)

    def _ckpt_rb(
        self, rb: ReplayBuffer | EnvIndependentReplayBuffer | EpisodeBuffer
    ) -> Tensor | Sequence[Tensor] | Sequence[Sequence[Tensor]]:
        """Modify the replay buffer in order to be consistent for the checkpoint.
        There could be 3 cases, depending on the buffers:

        1. The `ReplayBuffer` or `SequentialReplayBuffer`: a done is inserted in the last pos because the
            state of the environment is not saved in the checkpoint.
        2. The `EnvIndependentReplayBuffer`: for each buffer, the done in the last position is set to True
            (for the same reason of the point 1.).
        3. The `EpisodeBuffer`: the open episodes are discarded  because the
            state of the environment is not saved in the checkpoint.

        Args:
            rb (ReplayBuffer | EnvIndependentReplayBuffer | EpisodeBuffer): the buffer.

        Returns:
            The original state of the buffer.
        """
        if isinstance(rb, ReplayBuffer):
            # clone the true done
            state = rb["dones"][(rb._pos - 1) % rb.buffer_size, :].copy()
            # substitute the last done with all True values (all the environment are truncated)
            rb["dones"][(rb._pos - 1) % rb.buffer_size, :] = True
        elif isinstance(rb, EnvIndependentReplayBuffer):
            state = []
            for b in rb.buffer:
                state.append(b["dones"][(b._pos - 1) % b.buffer_size, :].copy())
                b["dones"][(b._pos - 1) % b.buffer_size, :] = True
        elif isinstance(rb, EpisodeBuffer):
            # remove open episodes from the buffer because the state of the environment is not saved
            state = rb._open_episodes
            rb._open_episodes = [[] for _ in range(rb.n_envs)]
        return state

    def _experiment_consistent_rb(
        self,
        rb: ReplayBuffer | EnvIndependentReplayBuffer | EpisodeBuffer,
        state: Tensor | Sequence[Tensor] | Sequence[Sequence[Tensor]],
    ):
        """Restore the state of the buffer consistent with the execution of the experiment.
        I.e., it undoes the changes in the _ckpt_rb function.

        Args:
            rb (ReplayBuffer | EnvIndependentReplayBuffer | EpisodeBuffer): the buffer.
            state (Tensor | Sequence[Tensor] | Sequence[Sequence[Tensor]]): the original state of the buffer.
        """
        if isinstance(rb, ReplayBuffer):
            # reinsert the true dones in the buffer
            rb["dones"][(rb._pos - 1) % rb.buffer_size, :] = state
        elif isinstance(rb, EnvIndependentReplayBuffer):
            for i, b in enumerate(rb.buffer):
                b["dones"][(b._pos - 1) % b.buffer_size, :] = state[i]
        elif isinstance(rb, EpisodeBuffer):
            # reinsert the open episodes to continue the training
            rb._open_episodes = state

    def _delete_old_checkpoints(self, ckpt_folder: pathlib.Path):
        ckpts = list(sorted(ckpt_folder.glob("*.ckpt"), key=os.path.getmtime))
        if len(ckpts) > self.keep_last:
            to_delete = ckpts[: -self.keep_last]
            [f.unlink() for f in to_delete]



from __future__ import annotations

import warnings
from math import isnan
from typing import Any, Dict, List, Optional, Tuple

import torch
from lightning import Fabric
from torch import Tensor
from torchmetrics import Metric


class MetricAggregatorException(Exception):
    """A custom exception used to report errors in use of timer class"""


class MetricAggregator:
    """A metric aggregator class to aggregate metrics to be tracked.
    Args:
        metrics (Optional[Dict[str, Metric]]): Dict of metrics to aggregate.
    """

    disabled: bool = False

    def __init__(self, metrics: Optional[Dict[str, Metric]] = None, raise_on_missing: bool = False):
        self.metrics: Dict[str, Metric] = {}
        if metrics is not None:
            self.metrics = metrics
        self._raise_on_missing = raise_on_missing

    def __iter__(self):
        return iter(self.metrics.keys())

    def add(self, name: str, metric: Metric):
        """Add a metric to the aggregator

        Args:
            name (str): Name of the metric
            metric (Metric): Metric to add.

        Raises:
            MetricAggregatorException: If the metric already exists.
        """
        if not self.disabled:
            if name not in self.metrics:
                self.metrics.setdefault(name, metric)
            else:
                if self._raise_on_missing:
                    raise MetricAggregatorException(f"Metric {name} already exists")
                else:
                    warnings.warn(
                        f"The key '{name}' is already in the metric aggregator. Nothing will be added.", UserWarning
                    )

    @torch.no_grad()
    def update(self, name: str, value: Any) -> None:
        """Update the metric with the value

        Args:
            name (str): Name of the metric
            value (Any): Value to update the metric with.

        Raises:
            MetricAggregatorException: If the metric does not exist.
        """
        if not self.disabled:
            if name not in self.metrics:
                if self._raise_on_missing:
                    raise MetricAggregatorException(f"Metric {name} does not exist")
                else:
                    warnings.warn(
                        f"The key '{name}' is missing from the metric aggregator. Nothing will be added.", UserWarning
                    )
            else:
                self.metrics[name].update(value)

    def pop(self, name: str) -> None:
        """Remove a metric from the aggregator with the given name
        Args:
            name (str): Name of the metric
        """
        if not self.disabled:
            if name not in self.metrics:
                if self._raise_on_missing:
                    raise MetricAggregatorException(f"Metric {name} does not exist")
                else:
                    warnings.warn(
                        f"The key '{name}' is missing from the metric aggregator. Nothing will be popped.", UserWarning
                    )
            self.metrics.pop(name, None)

    def reset(self):
        """Reset all metrics to their initial state"""
        if not self.disabled:
            for metric in self.metrics.values():
                metric.reset()

    def to(self, device: str | torch.device = "cpu") -> "MetricAggregator":
        """Move all metrics to the given device
        Args:
            device (str |torch.device, optional): Device to move the metrics to. Defaults to "cpu".
        """
        if not self.disabled:
            if self.metrics:
                for k, v in self.metrics.items():
                    self.metrics[k] = v.to(device)
        return self

    @torch.no_grad()
    def compute(self) -> Dict[str, Any]:
        """Reduce the metrics to a single value
        Returns:
            Reduced metrics
        """
        reduced_metrics = {}
        if not self.disabled:
            if self.metrics:
                for k, v in self.metrics.items():
                    reduced = v.compute()
                    is_tensor = torch.is_tensor(reduced)
                    if is_tensor and reduced.numel() == 1:
                        reduced_metrics[k] = reduced.item()
                    else:
                        if not is_tensor:
                            warnings.warn(
                                f"The reduced metric {k} is not a scalar tensor: type={type(reduced)}. "
                                "This may create problems during the logging phase.",
                                category=RuntimeWarning,
                            )
                        else:
                            warnings.warn(
                                f"The reduced metric {k} is not a scalar: size={v.size()}. "
                                "This may create problems during the logging phase.",
                                category=RuntimeWarning,
                            )
                        reduced_metrics[k] = reduced

                    is_tensor = torch.is_tensor(reduced_metrics[k])
                    if (is_tensor and torch.isnan(reduced_metrics[k]).any()) or (
                        not is_tensor and isnan(reduced_metrics[k])
                    ):
                        reduced_metrics.pop(k, None)
        return reduced_metrics


class RankIndependentMetricAggregator:
    def __init__(
        self,
        fabric: Fabric,
        metrics: Dict[str, Metric] | MetricAggregator,
    ) -> None:
        """This metric is useful when one wants to maintain per-rank-independent metrics of some quantities,
        while still being able to broadcast them to all the processes in a `torch.distributed` group. Internally,
        this metric uses a `MetricAggregator` to keep track of the metrics, and then broadcasts the metrics
        to all the processes thanks to Fabric.

        Args:
            fabric (Fabric): the fabric object.
            metrics (Sequence[str]): the metrics.
        """
        super().__init__()
        self._fabric = fabric
        self._aggregator = metrics
        if isinstance(metrics, dict):
            self._aggregator = MetricAggregator(metrics)
        for m in self._aggregator.metrics.values():
            m._to_sync = False
            m.sync_on_compute = False

    def update(self, name: str, value: float | Tensor) -> None:
        self._aggregator.update(name, value)

    @torch.no_grad()
    def compute(self) -> Tensor | Dict | List | Tuple:
        """Compute the means, one for every metric. The metrics are first broadcasted

        Returns:
            the computed metrics, broadcasted from and to every processes.
        """
        computed_metrics = self._aggregator.compute()
        gathered_data = self._fabric.all_gather(computed_metrics)
        return gathered_data

    def to(self, device: str | torch.device = "cpu") -> "RankIndependentMetricAggregator":
        """Move all metrics to the given device

        Args:
            device (str |torch.device, optional): Device to move the metrics to. Defaults to "cpu".
        """
        self._aggregator.to(device)
        return self

    def reset(self) -> None:
        """Reset the internal state of the metrics"""
        self._aggregator.reset()



"""Thank you @lorenzomammana: https://github.com/orobix/quadra/blob/main/quadra/utils/model_manager.py"""

from __future__ import annotations

import getpass
import os
import warnings
from abc import ABC, abstractmethod
from datetime import datetime
from typing import Any, Callable, Dict, Literal, Sequence, Set

import gymnasium as gym
import torch
from lightning import Fabric
from lightning.fabric.wrappers import _FabricModule
from lightning.pytorch.loggers.mlflow import MLFlowLogger

from sheeprl.utils.env import make_env
from sheeprl.utils.imports import _IS_MLFLOW_AVAILABLE

if not _IS_MLFLOW_AVAILABLE:
    raise ModuleNotFoundError(str(_IS_MLFLOW_AVAILABLE))

import mlflow  # noqa
from mlflow.entities import Run  # noqa
from mlflow.entities.model_registry import ModelVersion  # noqa
from mlflow.exceptions import RestException  # noqa
from mlflow.models.model import ModelInfo  # noqa
from mlflow.tracking import MlflowClient  # noqa

VERSION_MD_TEMPLATE = "## **Version {}**\n"
DESCRIPTION_MD_TEMPLATE = "### Description: \n{}\n"


class AbstractModelManager(ABC):
    """Abstract class for model managers."""

    @abstractmethod
    def __init__(self, fabric: Fabric) -> None:
        self.fabric = fabric

    @abstractmethod
    def register_model(
        self, model_location: str, model_name: str, description: str, tags: Dict[str, Any] | None = None
    ) -> Any:
        """Register a model in the model registry."""

    @abstractmethod
    def get_latest_version(self, model_name: str) -> Any:
        """Get the latest version of a model for all the possible stages or filtered by stage."""

    @abstractmethod
    def transition_model(self, model_name: str, version: int, stage: str, description: str | None = None) -> Any:
        """Transition the model with the given version to a new stage."""

    @abstractmethod
    def delete_model(self, model_name: str, version: int, description: str | None = None) -> None:
        """Delete a model with the given version."""

    @abstractmethod
    def register_best_models(
        self,
        experiment_name: str,
        models_info: Dict[str, Dict[str, Any]],
        metric: str = "Test/cumulative_reward",
        mode: Literal["max", "min"] = "max",
    ) -> Any:
        """Register the best models from an experiment."""

    @abstractmethod
    def download_model(self, model_name: str, version: int, output_path: str) -> None:
        """Download the model with the given version to the given output path."""


class MlflowModelManager(AbstractModelManager):
    """Model manager for Mlflow."""

    def __init__(self, fabric: Fabric, tracking_uri: str):
        if not _IS_MLFLOW_AVAILABLE:
            raise ModuleNotFoundError(str(_IS_MLFLOW_AVAILABLE))

        super().__init__(fabric)
        self.tracking_uri = tracking_uri

        mlflow.set_tracking_uri(self.tracking_uri)
        self.client = MlflowClient()

    def register_model(
        self, model_location: str, model_name: str, description: str | None = None, tags: Dict[str, Any] | None = None
    ) -> ModelVersion:
        """Register a model in the model registry.

        Args:
            model_location (str): The model uri.
            model_name (str): The name of the model after it is registered.
            description (str, optional): A description of the model, this will be added to the model changelog.
                Default to None.
            tags (Dict[str, Any], optional): A dictionary of tags to add to the model.
                Default to None.

        Returns:
            The model version.
        """
        model_version = mlflow.register_model(model_uri=model_location, name=model_name, tags=tags)
        self.fabric.print(f"Registered model {model_name} with version {model_version.version}")
        registered_model_description = self.client.get_registered_model(model_name).description

        if model_version.version == "1":
            header = "# MODEL CHANGELOG\n"
        else:
            header = ""

        new_model_description = VERSION_MD_TEMPLATE.format(model_version.version)
        new_model_description += self._get_author_and_date()
        new_model_description += self._generate_description(description)

        self.client.update_registered_model(model_name, header + registered_model_description + new_model_description)

        self.client.update_model_version(
            model_name, model_version.version, "# MODEL CHANGELOG\n" + new_model_description
        )

        return model_version

    def get_latest_version(self, model_name: str) -> ModelVersion:
        """Get the latest version of a model.

        Args:
            model_name (str): The name of the model.

        Returns:
            The model version.
        """
        latest_version = max(int(x.version) for x in self.client.get_latest_versions(model_name))
        model_version = self.client.get_model_version(model_name, latest_version)

        return model_version

    def transition_model(
        self, model_name: str, version: int, stage: str, description: str | None = None
    ) -> ModelVersion | None:
        """Transition a model to a new stage.

        Args:
            model_name (str): The name of the model.
            version (int): The version of the model
            stage (str): The stage of the model.
            description (str, optional): A description of the transition, this will be added to the model changelog.
                Default to None.
        """
        previous_stage = self._safe_get_stage(model_name, version)

        if previous_stage is None:
            return None

        if previous_stage.lower() == stage.lower():
            warnings.warn(f"Model {model_name} version {version} is already in stage {stage}")
            return self.client.get_model_version(model_name, version)

        self.fabric.print(f"Transitioning model {model_name} version {version} from {previous_stage} to {stage}")
        model_version = self.client.transition_model_version_stage(name=model_name, version=version, stage=stage)
        new_stage = model_version.current_stage
        registered_model_description = self.client.get_registered_model(model_name).description
        single_model_description = self.client.get_model_version(model_name, version).description

        new_model_description = "## **Transition:**\n"
        new_model_description += f"### Version {model_version.version} from {previous_stage} to {new_stage}\n"
        new_model_description += self._get_author_and_date()
        new_model_description += self._generate_description(description)

        self.client.update_registered_model(model_name, registered_model_description + new_model_description)
        self.client.update_model_version(
            model_name, model_version.version, single_model_description + new_model_description
        )

        return model_version

    def delete_model(self, model_name: str, version: int, description: str | None = None) -> None:
        """Delete a model.

        Args:
            model_name (str): The name of the model,
            version (int): The version of the model.
            description (str, optional): Why the model was deleted, this will be added to the model changelog.
                Default to None.
        """
        model_stage = self._safe_get_stage(model_name, version)

        if model_stage is None:
            return

        if (
            input(
                f"Model named `{model_name}`, version {version} is in stage {model_stage}, "
                "type the model name to continue deletion:"
            )
            != model_name
        ):
            warnings.warn("Model name did not match, aborting deletion")
            return

        self.fabric.print(f"Deleting model {model_name} version {version}")
        self.client.delete_model_version(model_name, version)

        registered_model_description = self.client.get_registered_model(model_name).description

        new_model_description = "## **Deletion:**\n"
        new_model_description += f"### Version {version} from stage: {model_stage}\n"
        new_model_description += self._get_author_and_date()
        new_model_description += self._generate_description(description)

        self.client.update_registered_model(model_name, registered_model_description + new_model_description)

    def register_best_models(
        self,
        experiment_name: str,
        models_info: Dict[str, Dict[str, Any]],
        metric: str = "Test/cumulative_reward",
        mode: Literal["max", "min"] = "max",
    ) -> Dict[str, ModelVersion] | None:
        """Register the best model from an experiment.

        Args:
            experiment_name (str): The name of the experiment.
            models_info (Dict[str, Dict[str, Any]]): A dictionary containing models information
                (path, description and tags).
            metric (str): The metric to use to determine the best model.
                Default to "Test/cumulative_reward".
            mode (Literal["max", "min"]): The mode to use to determine the best model, either "max" or "min".
                Defaulto to "max".

        Returns:
            The registered models version if successful, otherwise None.
        """
        if mode not in ["max", "min"]:
            raise ValueError(f"Mode must be either 'max' or 'min', got {mode}")

        experiment_id = self.client.get_experiment_by_name(experiment_name).experiment_id
        runs = self.client.search_runs(experiment_ids=[experiment_id])

        if len(runs) == 0:
            self.fabric.print(f"No runs found for experiment {experiment_name}")
            return None

        best_run: Run | None = None
        best_run_artifacts: Sequence[str] | Set[str] | None = None
        models_path = [v["path"] for v in models_info.values()]
        for run in runs:
            run_artifacts = [x.path for x in self.client.list_artifacts(run.info.run_id) if x.path in models_path]

            if len(run_artifacts) == 0 or run.data.metrics.get(metric) is None:
                # If we don't find the given model path, skip this run
                # If the run has not the target metric, skip this run
                continue

            if best_run is None:
                best_run = run
                best_run_artifacts = set(run_artifacts)
                continue
            if mode == "max":
                if run.data.metrics[metric] > best_run.data.metrics[metric]:
                    best_run = run
            else:
                if run.data.metrics[metric] < best_run.data.metrics[metric]:
                    best_run = run

        if best_run is None:
            self.fabric.print(f"No runs found for experiment {experiment_name} with the given metric")
            return None

        models_version = {}
        for k, v in models_info.items():
            if v["path"] in best_run_artifacts:
                best_model_uri = f"runs:/{best_run.info.run_id}/{v['path']}"
                models_version[k] = self.register_model(
                    model_location=best_model_uri, model_name=v["name"], tags=v["tags"], description=v["description"]
                )

        return models_version

    def download_model(self, model_name: str, version: int, output_path: str) -> None:
        """Download the model with the given version to the given output path.

        Args:
            model_name (str): The name of the model.
            version (int): The version of the model.
            output_path (str): The path to save the model to.
        """
        artifact_uri = self.client.get_model_version_download_uri(model_name, version)
        self.fabric.print(f"Downloading model {model_name} version {version} from {artifact_uri} to {output_path}")
        if not os.path.exists(output_path):
            self.fabric.print(f"Creating output path {output_path}")
            os.makedirs(output_path)
        mlflow.artifacts.download_artifacts(artifact_uri=artifact_uri, dst_path=output_path)

    @staticmethod
    def _generate_description(description: str | None = None) -> str:
        """Generate the description markdown template."""
        if description is None:
            return ""

        return DESCRIPTION_MD_TEMPLATE.format(description)

    @staticmethod
    def _get_author_and_date() -> str:
        """Get the author and date markdown template."""
        author_and_date = f"### Author: {getpass.getuser()}\n"
        author_and_date += f"### Date: {datetime.now().astimezone().strftime('%d/%m/%Y %H:%M:%S %Z')}\n"

        return author_and_date

    def _safe_get_stage(self, model_name: str, version: int) -> str | None:
        """Get the stage of a model version.

        Args:
            model_name (str): The name of the model.
            version (int): The version of the model

        Returns:
            The stage of the model version if it exists, otherwise None.
        """
        try:
            model_stage = self.client.get_model_version(model_name, version).current_stage
            return model_stage
        except RestException:
            self.fabric.print(f"Model named {model_name} with version {version} does not exist")
            return None


def register_model_from_checkpoint(
    fabric: Fabric,
    cfg: Dict[str, Any],
    state: Dict[str, Any],
    log_models_from_checkpoint: Callable[
        [Fabric, gym.Env | gym.Wrapper, Dict[str, Any], Dict[str, Any]], Dict[str, ModelInfo]
    ],
):
    tracking_uri = getattr(cfg, "tracking_uri", None) or os.getenv("MLFLOW_TRACKING_URI", None)
    if tracking_uri is None:
        raise ValueError(
            "The tracking uri is not defined, use an mlflow logger with a tracking uri or define the "
            "MLFLOW_TRACKING_URI environment variable."
        )
    # Creating the environment for agent instantiation
    env = make_env(
        cfg,
        cfg.seed,
        0,
        None,
        "test",
        vector_env_idx=0,
    )()
    observation_space = env.observation_space
    if not isinstance(observation_space, gym.spaces.Dict):
        raise RuntimeError(f"Unexpected observation type, should be of type Dict, got: {observation_space}")
    if cfg.algo.cnn_keys.encoder + cfg.algo.mlp_keys.encoder == []:
        raise RuntimeError(
            "You should specify at least one CNN keys or MLP keys from the cli: "
            "`cnn_keys.encoder=[rgb]` or `mlp_keys.encoder=[state]`"
        )

    mlflow.set_tracking_uri(tracking_uri)
    # If the user does not specify the experiment, than, create a new experiment
    if cfg.run.id is None and cfg.experiment.id is None:
        cfg.experiment.id = mlflow.create_experiment(cfg.experiment.name)
    # Log the models
    models_info = log_models_from_checkpoint(fabric, env, cfg, state)
    model_manager = MlflowModelManager(fabric, tracking_uri)
    if not set(cfg.model_manager.models.keys()).issubset(models_info.keys()):
        raise RuntimeError(
            f"The models you want to register must be a subset of the models of the {cfg.algo.name} agent. "
            f"{len(cfg.model_manager.models)} model registration "
            f"configs are given, but the agent has {len(models_info)} models. "
            f"\nModels specified in the configs: {cfg.model_manager.models.keys()}."
            f"\nModels of the {cfg.algo.name} agent: {cfg.model_manager.models.keys()}."
        )
    # Register the models specified in the configs
    for k, cfg_model in cfg.model_manager.models.items():
        model_manager.register_model(
            models_info[k]._model_uri, cfg_model["model_name"], cfg_model["description"], cfg_model["tags"]
        )


def register_model(
    fabric: Fabric,
    log_models: Callable[
        [Dict[str, Any], Dict[str, torch.nn.Module | _FabricModule], str, str | None, str | None], Dict[str, ModelInfo]
    ],
    cfg: Dict[str, Any],
    models_to_log: Dict[str, torch.nn.Module | _FabricModule],
):
    if len(fabric.loggers) == 0:
        raise RuntimeError("No logger is defined, try to set the `metric.log_level=1` from the CLI")
    elif len(fabric.loggers) > 0 and not isinstance(fabric.logger, MLFlowLogger):
        raise RuntimeError(
            "The logger is not an mlflow logger, try to set the `logger@metric.logger=mlflow` from the CLI."
        )
    tracking_uri = getattr(fabric.logger, "_tracking_uri", None) or os.getenv("MLFLOW_TRACKING_URI", None)
    if tracking_uri is None:
        raise ValueError(
            "The tracking uri is not defined, use an mlflow logger with a tracking uri or define the "
            "MLFLOW_TRACKING_URI environment variable."
        )
    cfg_model_manager = cfg.model_manager
    # Retrieve run_id, if None, create a new run
    run_id = None
    if len(fabric.loggers) > 0:
        run_id = getattr(fabric.logger, "run_id", None)
    mlflow.set_tracking_uri(tracking_uri)
    experiment_id = None
    run_name = None
    if run_id is None:
        experiment = mlflow.get_experiment_by_name(cfg.exp_name)
        experiment_id = mlflow.create_experiment(cfg.exp_name) if experiment is None else experiment.experiment_id
        run_name = f"{cfg.algo.name}_{cfg.env.id}_{datetime.today().strftime('%Y-%m-%d %H:%M:%S')}"
    models_info = log_models(cfg, models_to_log, run_id, experiment_id, run_name)
    model_manager = MlflowModelManager(fabric, tracking_uri)
    if len(models_info) != len(cfg_model_manager.models):
        raise RuntimeError(
            f"The number of models of the {cfg.algo.name} agent must be equal to the number "
            f"of models you want to register. {len(cfg_model_manager.models)} model registration "
            f"configs are given, but the agent has {len(models_info)} models."
        )
    for k, cfg_model in cfg_model_manager.models.items():
        model_manager.register_model(
            models_info[k]._model_uri, cfg_model["model_name"], cfg_model["description"], cfg_model["tags"]
        )
